"""
Models an adaptive optics observation system in as much detail required to get a PSF.
"""
import sys

import typing
from typing import Callable

import numpy as np
import scipy as sp
import scipy.ndimage
import scipy.interpolate
import scipy.signal
import inspect

import matplotlib.pyplot as plt

import aopp_deconv_tool.cfg.logs
import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.array

from aopp_deconv_tool.geo_array import GeoArray, plot_ga
from aopp_deconv_tool.optics.function import PointSpreadFunction, OpticalTransferFunction, PhasePowerSpectralDensity
from aopp_deconv_tool.instrument_model.instrument_base import InstrumentBase


_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'INFO')


def downsample(a, s):
	return sp.signal.convolve(a, np.ones([s]*a.ndim)/(s**a.ndim), mode='valid')[tuple(slice(None,None,s) for _ in a.shape)]





class PSFModel:
	"""
	Class that calculates a model point spread function.
	"""
	def __init__(self,
			telescope_otf_model : 
				Callable[[tuple[int,...], float,float,...],OpticalTransferFunction]
				| OpticalTransferFunction, # axes in units of wavelength/rho
			atmospheric_turbulence_psd_model : 
				Callable[[np.ndarray,...],PhasePowerSpectralDensity]
				| PhasePowerSpectralDensity, # takes in spatial scale axis, returns phase power spectrum distribution
			adaptive_optics_psd_model : 
				Callable[[np.ndarray, ...], PhasePowerSpectralDensity]
				| PhasePowerSpectralDensity, # takes in spatial scale axis, returns phase PSD
			instrument : InstrumentBase # instrument used to take observation, defines certain scale parameters
		):
	
		
		if callable(telescope_otf_model):
			self.telescope_otf_model = telescope_otf_model
			self.telescope_otf = None
		else:
			self.telescope_otf_model = None
			self.telescope_otf = telescope_otf_model
		
		if callable(atmospheric_turbulence_psd_model):
			self.atmospheric_turbulence_psd_model = atmospheric_turbulence_psd_model
			self.atmospheric_turbulence_psd = None
		else:
			self.atmospheric_turbulence_psd_model = None
			self.atmospheric_turbulence_psd = atmospheric_turbulence_psd_model
	
		if callable(adaptive_optics_psd_model):
			self.adaptive_optics_psd_model = adaptive_optics_psd_model
			self.adaptive_optics_psd = None
		else:
			self.adaptive_optics_psd_model=None
			self.adaptive_optics_psd = adaptive_optics_psd_model
		
		
		self.instrument = instrument
		self.specific_model_ready = False
		
		
	def ao_corrections_to_phase_psd(self, 
			phase_psd, 
			ao_phase_psd, 
			f_ao, 
			ao_correction_amplitude=1, 
			ao_correction_frac_offset=0
		):
		"""
		Apply adaptive optics corrections to the phase power spectrum distribution of the atmosphere
		"""
		if ao_phase_psd is None: 
			return phase_psd.copy()
		
		
		f_delta = 1.5
		
		_lgr.debug(f'{phase_psd.data=} {ao_phase_psd.data=}')
		_lgr.debug(f'{f_ao=} {ao_correction_amplitude=} {ao_correction_frac_offset=}')
		
		f_mesh = phase_psd.mesh
		f_mag = np.sqrt(np.sum(f_mesh**2, axis=0))
		
		_lgr.debug(f'{np.max(f_mag)=}')
		_lgr.debug(f'{np.count_nonzero(np.isnan(f_mesh))=} {np.count_nonzero(np.isnan(f_mag))=}')
		
		f_ao_correct = f_mag <= f_ao
		f_ao_continuity_region = ((f_ao-f_delta) <= f_mag) & (f_mag < (f_ao+f_delta))
		
		_lgr.debug(f'{np.count_nonzero(f_ao_continuity_region)=} {np.sum(f_ao_continuity_region)=}')
		
		f_ao_continuity_factor = np.nanmean(phase_psd.data[f_ao_continuity_region])
		f_ao_correction_offset = np.nanmean(ao_phase_psd.data[f_ao_continuity_region])
		
		_lgr.debug(f'{f_ao_continuity_factor=} {f_ao_correction_offset=}')
		
		
		ao_corrected_psd = np.array(phase_psd.data)
		ao_corrected_psd[f_ao_correct] = (
			np.exp(ao_correction_amplitude)*(ao_phase_psd.data[f_ao_correct] - f_ao_correction_offset)
			+ np.exp(ao_correction_frac_offset)
		)*f_ao_continuity_factor
		
		return PhasePowerSpectralDensity(ao_corrected_psd, phase_psd.axes)
	
	
	def optical_transfer_fuction_from_phase_psd(self, phase_psd, mode='classic', s_factor=0):
		"""
		From a phase power spectral distribution, derive an optical transfer function.
		"""
		phase_autocorr = phase_psd.ifft()
		phase_autocorr.data = phase_autocorr.data.real
		
		match mode:
			case "classic":
				centre_point = tuple(_s//2 for _s in phase_autocorr.data.shape)
				otf_data = np.exp(phase_autocorr.data - phase_autocorr.data[centre_point])
				
				# Work out the offset from zero of the OTF
				centre_idx_offsets = nph.array.offsets_from_point(otf_data.shape)
				centre_idx_dist = np.sqrt(np.sum(centre_idx_offsets**2, axis=0))
				outer_region_mask = centre_idx_dist > (centre_idx_dist.shape[0]//2)*0.9
				otf_data_offset_from_zero = np.sum(otf_data[outer_region_mask])/np.count_nonzero(outer_region_mask)
				
				_lgr.debug(f'{otf_data_offset_from_zero=}')
				# Subtract the offset to remove the delta-function spike
				otf_data -= otf_data_offset_from_zero 
				# If required, add back on a fraction of the maximum to add part of the spike back in.
				otf_data += s_factor*np.max(otf_data)
				
				otf = GeoArray(otf_data, phase_autocorr.axes)
	
			case "adjust":
				otf = GeoArray(np.abs(phase_autocorr.data), phase_autocorr.axes)
		return otf


	def __call__(self,
			telescope_otf_model_args,
			atmospheric_turbulence_psd_model_args,
			adaptive_optics_psd_model_args,
			f_ao,
			ao_correction_amplitude=1,
			ao_correction_frac_offset=0,
			s_factor=0,
			mode='adjust'
		):
		"""
		Calculate psf in terms of rho/wavelength for given parameters. 
		We are making an implicit assumption that all of the calculations can be
		done in rho/wavelength units.
		"""
		
		_lgr.debug(f'{telescope_otf_model_args=}')
		_lgr.debug(f'{atmospheric_turbulence_psd_model_args=}')
		_lgr.debug(f'{adaptive_optics_psd_model_args=}')
		_lgr.debug(f'{f_ao=}')
		_lgr.debug(f'{ao_correction_amplitude=}')
		_lgr.debug(f'{ao_correction_frac_offset=}')
		_lgr.debug(f'{s_factor=}')
		_lgr.debug(f'{mode=}')
		
		self.shape = self.instrument.obs_shape
		self.expansion_factor = self.instrument.expansion_factor
		self.supersample_factor = self.instrument.supersample_factor
		self.adaptive_optics_psd_model_args = adaptive_optics_psd_model_args
		
		
		self.telescope_otf_model_args = telescope_otf_model_args
		self.atmospheric_turbulence_psd_model_args = atmospheric_turbulence_psd_model_args
		self.f_ao = f_ao
		self.ao_correction_amplitude = ao_correction_amplitude
		self.ao_correction_frac_offset = ao_correction_frac_offset
		self.mode= mode
		self.s_factor = s_factor
		
		self.specific_model_ready = True
		return self
	
	
	
	def at(self, wavelength, plots=True):
		"""
		Calculate psf for a given angular scale and wavelength, i.e., convert
		from rho/wavelength units to rho units.
		"""
		if not self.specific_model_ready:
			raise RuntimeError(f'{type(self).__name__}.at({wavelength}) cannot be called yet. You need to specify model parameters with {type(self).__name__}.__call__(...) first.')
		
		wav_factor = (wavelength/self.instrument.ref_wavelength)
		ang_wav_factor = (self.instrument.obs_pixel_size/wavelength)
		_lgr.debug(f'{wav_factor=} {ang_wav_factor=}')
		
		# Get the telescope optical transfer function
		if self.telescope_otf_model is not None:
			self.telescope_otf = self.telescope_otf_model(self.shape, self.expansion_factor, self.supersample_factor, *self.telescope_otf_model_args)
		
		telescope_otf = self.telescope_otf#GeoArray(self.telescope_otf.data, tuple(a*ang_wav_factor for a in self.telescope_otf.axes))
				
		_lgr.debug(f'{telescope_otf.data.shape=} {tuple(x.size for x in telescope_otf.axes)=}')
		_lgr.debug(f'{telescope_otf.axes=}')
		
		if plots: plot_ga(telescope_otf, lambda x: np.log(np.abs(x)), 'diffraction limited otf', 'arbitrary units', 'wavelength/rho')
		if plots: plot_ga(telescope_otf.ifft(), lambda x: np.log(np.abs(x)), 'diffraction limited psf', 'arbitrary units', 'rho/wavelength')
		
		_lgr.debug(f'{self.instrument.obs_scale=} {self.instrument.obs_shape=}')
		
		#f_axes = telescope_otf.ifft().axes
		f_axes = tuple(x/wav_factor for x in telescope_otf.ifft().axes) # TEST THIS
		#f_axes = tuple(a/self.instrument.obs_pixel_size for a,sc,sh in zip(self.telescope_otf.ifft().axes,self.instrument.obs_scale,self.instrument.obs_shape))
		
		_lgr.debug(f'{tuple(x.size for x in f_axes)=}')
		_lgr.debug(f'{f_axes=}')
		
		# Get the atmospheric phase power spectral density
		if self.atmospheric_turbulence_psd_model is not None:
			self.atmospheric_turbulence_psd = self.atmospheric_turbulence_psd_model(
				f_axes,
				wavelength,
				*self.atmospheric_turbulence_psd_model_args
			)
		if plots: plot_ga(self.atmospheric_turbulence_psd, lambda x: np.log(np.abs(x)), 'atmospheric_turbulence_psd', 'arbitrary units', 'rho/wavelength')
		
		# Get the adaptive optics phase power spectral density
		if self.adaptive_optics_psd_model is not None:
			self.adaptive_optics_psd = self.adaptive_optics_psd_model(
				f_axes, 
				*self.adaptive_optics_psd_model_args
			)
		if plots: plot_ga(self.adaptive_optics_psd, lambda x: np.log(np.abs(x)), 'adaptive_optics_psd', 'arbitrary units', 'rho/wavelength')
		_lgr.debug(f'{self.atmospheric_turbulence_psd.data=}')
		_lgr.debug(f'{self.adaptive_optics_psd.data=}')
		
		
		
		
		
		# Apply the adapative optics phase power spectral density corrections to the atmospheric phase power spectral density
		ao_corrected_atm_phase_psd = self.ao_corrections_to_phase_psd(
			self.atmospheric_turbulence_psd, 
			self.adaptive_optics_psd, 
			self.f_ao,#*wav_factor, # TEST THIS
			self.ao_correction_amplitude,#*wav_factor, # TEST THIS
			self.ao_correction_frac_offset
		)
		if plots: plot_ga(ao_corrected_atm_phase_psd, lambda x: np.log(np.abs(x)), 'ao_corrected_atm_phase_psd', 'arbitrary units', 'rho/wavelength')
		
		ao_corrected_otf = self.optical_transfer_fuction_from_phase_psd(ao_corrected_atm_phase_psd, self.mode, self.s_factor)
		if plots: plot_ga(ao_corrected_otf, lambda x: np.log(np.abs(x)), 'ao_corrected_otf', 'arbitrary units', 'wavelength/rho')
		if plots: plot_ga(ao_corrected_otf.ifft(), lambda x: np.log(np.abs(x)), 'ao_corrected_psf', 'arbitrary units', 'rho/wavelength')
		
		_lgr.debug(f'{ao_corrected_otf.data.shape=} {telescope_otf.data.shape=}')
		_lgr.debug(f'{ao_corrected_otf.data=}')
		_lgr.debug(f'{telescope_otf.data=}')
		
		"""
		t_otf_wav_axes = tuple(a/ang_wav_factor for a in self.telescope_otf.axes)
		interp = sp.interpolate.RegularGridInterpolator(
			self.telescope_otf.axes, 
			self.telescope_otf.data,method='linear',
			bounds_error=False, 
			fill_value=np.min(self.telescope_otf.data)
		)
		points = np.swapaxes(np.array(np.meshgrid(*t_otf_wav_axes)), 0,-1)
		t_otf_wav = GeoArray(np.array(interp(points)), t_otf_wav_axes)
		if plots: plot_ga(t_otf_wav, lambda x: np.log(np.abs(x)), 't_otf_wav', 'arbitrary units', '1/rho')
		if plots: plot_ga(t_otf_wav.ifft(), lambda x: np.log(np.abs(x)), 't_psf_wav', 'arbitrary units', 'rho')
		"""
		
		
		# Combination of diffraction-limited optics, atmospheric effects, AO correction to atmospheric effects
		otf_full = GeoArray(
			ao_corrected_otf.data * telescope_otf.data, 
			#(ao_corrected_otf.data/np.nansum(ao_corrected_otf.data)) * (self.telescope_otf.data/np.nansum(self.telescope_otf.data)), 
			telescope_otf.axes
		)
		if plots: plot_ga(otf_full, lambda x: np.log(np.abs(x)), 'otf full', 'arbitrary units', 'wavelength/rho')
		
		psf_full = otf_full.ifft()
		if plots: plot_ga(psf_full, lambda x: np.log(np.abs(x)), 'psf full', 'arbitrary units', 'rho/wavelength')
		
		psf_full = PointSpreadFunction(np.abs(psf_full.data), psf_full.axes)
		_lgr.debug(f'{psf_full.data=}')
		

		
		output_axes = tuple(np.linspace(-s/2*ang_wav_factor,s/2*ang_wav_factor,s*self.instrument.supersample_factor) for s in self.shape)
		
		#output_axes = tuple(np.linspace(-z/2,z/2,s) for z,s in zip(self.instrument.obs_scale*self.instrument.ref_wavelength,self.psf_full.data.shape))
		_lgr.debug(f'{output_axes=}')
		
		rho_axes = psf_full.axes#tuple(a*wavelength for a in self.psf_full.axes)
		_lgr.debug(f'{rho_axes=}')
		
		# swap output and rho axes to see if it makes a difference
		#temp = output_axes
		#output_axes = rho_axes
		#rho_axes = temp
				
		interp = sp.interpolate.RegularGridInterpolator(rho_axes, psf_full.data,method='linear', bounds_error=False, fill_value=np.min(psf_full.data))
		
		points = np.swapaxes(np.array(np.meshgrid(*output_axes)), 0,-1)
		
		
		
		sample_size = tuple(s1/(self.instrument.expansion_factor*s2) for s1,s2 in zip(psf_full.data.shape, self.shape))
		assert all([s==int(sample_size[0]) for s in sample_size]), "Should have a constant integer factor"
		sample_size = int(sample_size[0])
		_lgr.debug(f'{sample_size=}')
		
		result = GeoArray(downsample(np.array(interp(points)), sample_size), tuple(a[::sample_size] for a in output_axes))
		
		#result = GeoArray(downsample(self.psf_full.data, sample_size), tuple(wav_factor*a[::sample_size] for a in self.psf_full.axes))
		_lgr.debug(f'{result.data.shape=}')
		
		_lgr.debug(f'{result.data=}')
		
		if plots: plot_ga(result, lambda x: np.log(np.abs(x)), f'psf at {wavelength=}', 'arbitrary units', 'radians')
		
		return result









if __name__=='__main__':
	from geometry.shape import Circle
	from optics.geometric.optical_component import Aperture, Obstruction,Refractor
	from optics.telescope_model import optical_transfer_function_of_optical_component_set
	from optics.turbulence_model import phase_psd_von_karman_turbulence
	from optics.adaptive_optics_model import phase_psd_fetick_2019_moffat_function
	from instrument_model.vlt import VLT
	
	
	
	instrument = VLT.muse()
	
	psf_model = PSFModel(
		instrument.optical_transfer_function(3,1),
		phase_psd_von_karman_turbulence,
		phase_psd_fetick_2019_moffat_function,
	)
	
	
	psf = psf_model(
		instrument.obs_shape, 
		instrument.expansion_factor, 
		instrument.supersample_factor, 
		instrument.f_ao, 
		None, 
		(	0.17, 
			2, 
			8
		), 
		(	instrument.f_ao,
			np.array([5E-2,5E-2]),#np.array([5E-2,5E-2]),
			1.6,#1.6
			2E-2,#2E-3
			0.05,#0.05
		),
		plots=False
	)
	
	psf_model.at(tuple(101*x for x in (1.212E-7, 1.212E-7)), 5E-7)
	psf_model.at(tuple(101*x for x in (1.212E-7, 1.212E-7)), 6E-7)
	psf_model.at(tuple(101*x for x in (1.212E-7, 1.212E-7)), 7E-7)
	psf_model.at(tuple(101*x for x in (1.212E-7, 1.212E-7)), 8E-7)
	



"""
Routine for parsing a string containing the path, extension, slices, and axes
of a FITS file that we want to operate on.
"""
import os
from pathlib import Path
from collections import namedtuple 
import dataclasses as dc
from astropy.io import fits


import aopp_deconv_tool.text as text
import aopp_deconv_tool.cast as cast

import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.slice

import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.header

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')

HDUIdentifier = namedtuple('HDUIdentifier', ('index', 'name', 'version'))


def HDUIdentifiers_from(hdul : fits.HDUList) -> tuple[HDUIdentifier]:
	return tuple(HDUIdentifier(i, hdu.name, hdu.ver) for i, hdu in enumerate(hdul))
	


@dc.dataclass(init=True, repr=True, order=False, eq=False, slots=True)
class FitsInfo:
	path : Path = dc.field(init=True, repr=True, hash=False, compare=False)
	_n_hdus : int = dc.field(default=None, init=False, repr=True, hash=False, compare=False)
	_hdu_identifiers : tuple[HDUIdentifier] = dc.field(default=None, init=False, repr=True, hash=False, compare=False)

	def refresh(self):
		with fits.open(self.path) as hdul:
			self._hdu_identifiers = HDUIdentifiers_from(hdul)
			self._n_hdus = len(self._hdu_identifiers)
	
	@property
	def n_hdus(self):
		if self._n_hdus is None:
			self.refresh()
		return self._n_hdus
		
	@property
	def hdu_identifiers(self):
		if self._hdu_identifiers is None:
			self.refresh()
		return self._hdu_identifiers
		

FitsSpecifier = namedtuple('FitsSpecifier', ('path', 'ext', 'slices', 'axes'))



AxesInfo = namedtuple('AxesInfo', ('description', 'default_callable'))
axes_type_info={
	"SPECTRAL" : AxesInfo("wavelength/frequency varies along this axis", aph.fits.header.get_spectral_axes),
	"CELESTIAL" : AxesInfo("sky position varies along this axis", aph.fits.header.get_celestial_axes),
	"POLARISATION" : AxesInfo("polarisation varies along this axis, could be linear of circular", aph.fits.header.get_polarisation_axes),
	"TIME" : AxesInfo("time varies along this axis", aph.fits.header.get_time_axes),
	"ALL" : AxesInfo("all axes of a fits file", aph.fits.header.get_all_axes),
}

help_fmt = """\
# FITS SPECIFIER #

	A string that describes which FITS file to load, the extension (i.e backplane) to use, the slices (i.e., sub-regions)
	that should be operated upon, and (if required) the semantic meaning of the data axes.

	Format:
		.../path/to/a.fits{{ext}}[slice0,slice1,...]{{axes_type_1:(ax11,ax12,...),axes_type_2:(ax21,ax22,...)}}
		
		Where:
			.../path/to/a.fits
				A path to a FITS file. Must be present.
			ext : str | int
				Fits extension, either a string or an index. If not present, will assume PRIMARY hdu, (hdu index 0)
			sliceM : str
				Slice of Mth axis in the normal python `start:stop:step` format. If not present will assume slice(None) for an axis.
			axes_type_N : axes_type
				Type of the Nth axes set. Axes types are detailed below. They are used to tell a program what to do with
				an axis when normal FITS methods of description fail (or as a backup). Note, the type (and enclosing curly backets
				`{{}}` can be ommited if a program only accepts one or two axes types. In that case, the specified axes will be
				assumed to be of the first type specified in the documentation, and the remaining axes of the second type (if there
				is a second type).
			axNL : int
				Axes of the extension that are of the specified type. If not present, will assume all axes are of the first type
				specified in the documentation.
		
		Accepted axes_type:
			{axes_types}
			
		NOTE: As the format for a FITS specifier uses characters that a terminal application may interpret as special characters, 
			e.g., square/curly/round brackets, and colons. It can be better to wrap specifiers in quotes or single quotes. When 
			doing this, it is important to un-escape any previously escaped characters. 
			For example, specifies with timestamps in them would normally have the colons escaped, but when wrapped in quotes 
			this is not required. E.g., the specifier ./example_data/MUSE.2019-10-17T23\\:46\\:14.117_normalised.fits(1,2) will not 
			play nice with the bash shell due to the brackets. However, wrapping it in single quotes and removing the escaping 
			slashes from the colons means it will work. E.g., './example_data/MUSE.2019-10-17T23:46:14.117_normalised.fits(1,2)'
		
		Examples:
			~/home/datasets/MUSE/neptune_obs_1.fits{{DATA}}[100:200,:,:](1,2)
				Selects the "DATA" extension, slices the 0th axis from 100->200 leaving the others untouched, and signals that axes 1 and 2 are important e.g., they are the RA-DEC axes.
			
			~/home/datasets/MUSE/neptune_obs_1.fits{{DATA}}[100:200](1,2)
				Does the same thing as above, but omits un-needed slice specifiers.
			
			~/home/datasets/MUSE/neptune_obs_1.fits{{DATA}}[100:200]{{CELESTIAL:(1,2)}}
				Again, same as above, but adds explicit axes type.
"""

def get_help(axes_types : list[str]):
	"""
	Generates the help string for a list of axes_types.
	"""
	for x in axes_types:
		if x not in axes_type_info:
			raise RuntimeError(f"axes_type '{x}' is not one of the known axes_types {list(axes_type_info.keys())}")
	
	return help_fmt.format(axes_types='\n\t\t\t'.join([k+'\n\t\t\t\t'+axes_type_info[k].description for k in axes_types]))



def parse_axes_type_list(axes_type_list : str, axes_types: list[str] | tuple[str]):
	"""
	Parses the axes type list at the end of the specifier string.
	
	Arguments:
		axes_type_list : str
			"axes_type_1:(ax11,ax22,...),axes_type_2:(ax21,ax22,...)" or "(ax11,ax22)"
		axes_types : list[str] | tuple[str]
			List of acceptable axes types
	"""
	
	# first split on commas
	axes = {}
	#_lgr.debug(f"{text.split_around_brackets(axes_type_list)=}")
	for i, axes_type_str in enumerate(text.split_around_brackets(axes_type_list)):
		# axes_type_str = "axes_type_1:(ax11,ax12,...)" or "(ax11,ax12,...)"
		#_lgr.debug(f'{axes_type_str=}')
		n_colon = axes_type_str.count(':')
		if n_colon == 0:
			axtype, axtuple_str = (axes_types[i], axes_type_str)
		elif n_colon == 1:
			axtype, axtuple_str = axes_type_str.split(':')
		else:
			raise RuntimeError("axes_type_str '{axes_type_str}' has more than one colon, it is malformed.")
		
		# now parse axtuple_str
		assert axtuple_str[0] == '(' and axtuple_str[-1] ==')', f"axes tuple '{axtuple_str}' must be enclosed in brackets"
		
		axtuple = text.to_tuple(axtuple_str[1:-1], int)
		axes[axtype] = axtuple
	
	return(axes)




	

def parse(specifier : str, axes_types : list[str]):
	f"""
	Parses a string that specifies a fits file, extension, and slices.
	
	# ARGUMENTS #
		specifier : str
			A FITS specifier string, has the following format:
				{text.indent(get_help(axes_types), 4)}
		axes_types : list[str]
			A list of axes types that we expect. Should be a list of entries from `axes_type_info.keys()`
	
	# RETURNS #
		FitsSpecifier(path, ext, slices, axes)
			
	"""
	help_string = get_help(axes_types)
	
	# Go through cases from end to start
	
	try:
		# get axes information
		j = len(specifier) - 1
		axes_type_list = None
		if specifier[j-1:] == ')}':
			# we have axes_types in the specifier.
			i = specifier.rfind('{',0,j)
			if i == -1:
				raise RuntimeError("FITS specifier is malformed, missing '{'")
			axes_type_list = specifier[i+1:j]
			specifier = specifier[:i]
			j = i-1
		elif specifier[j] == ')':
			i = specifier.rfind('(',0,j)
			if i == -1:
				raise RuntimeError("FITS specifier is malformed, missing '('")
			axes_type_list = specifier[i:j+1]
			specifier = specifier[:i]
			j = i-1
		#_lgr.debug(f'{axes_type_list=}')	
		if axes_type_list is not None:
			axes = parse_axes_type_list(axes_type_list, axes_types)
		else:
			axes = None
		_lgr.debug(f'{axes=}')
		#print(f'{axes=}')
		if axes is not None:
			for k,v in axes.items():
				if len(v) == 0:
					raise RuntimeError(f'Could not work out axes "{k}" automatically, please specify axes explicitly')
		#else:
		#	raise RuntimeError(f'Could not work out axes automatically, please specify axes explicitly')
			
		
		# get slice information
		if specifier[j] == ']':
			# we have slice information
			i = specifier.rfind('[',0,j)
			if i == -1:
				raise RuntimeError("FITS specifier is malformed, missing '['")
			slice_tuple_str = specifier[i+1:j] # exclude "[" and "]" to give "a:b:c,d:e:f,..."
			slices = nph.slice.from_string(slice_tuple_str)
			specifier = specifier[:i]
			j = i-1
		else:
			slices=None
		
		
		# get extension
		if specifier[j] == '}':
			i = specifier.rfind('{',0,j)
			if i == -1:
				raise RuntimeError("FITS specifier is malformed, missing '{'")
			ext_str = specifier[i+1:j]
			ext = cast.to_any(ext_str, (int, str))
			specifier = specifier[:i]
			j = i-1
		else:
			ext = None # extension was not specified, determine from file later
		
	except Exception as e:
		e.add_note(help_string)
		raise
	
	# get path
	path = specifier
	
	if not os.path.isfile(path):
		raise FileNotFoundError(f'file "{path}" not found')
	
	fits_info = FitsInfo(path)
	
	# Set default extension if one was not specified
	if ext is None:
		hdu_names = tuple(x.name for x in fits_info.hdu_identifiers)
		hdu_name_priority_list = ('SCI', 'DATA', 'PRIMARY')
		for x in hdu_name_priority_list:
			#print(f'{x=} {hdu_names=} {fits_info=}')
			if x in hdu_names:
				hdu_ident = fits_info.hdu_identifiers[hdu_names.index(x)]
				ext = 'PRIMARY' if hdu_ident.index==0 else hdu_ident.name
				break
		# If we didn't find any names in our list, use the frist one
		if ext is None:
			ext = 0
	

	if slices is None or axes is None:
		try:
			hdr = fits.getheader(path,ext=ext if type(ext) is int else (ext,1))
		except Exception as e:
			pass
		else:
			if slices is None:
				slices = tuple(slice(None) for _ in range(hdr['NAXIS']))
			else:
				#slices = tuple(slices[i] if (i < len(slices)) else slice(None) for i in range(hdr['NAXIS']))
				pass
				
			if axes is None:
				axes = {}
				for ax_name in axes_types:
					axes[ax_name] = axes_type_info[ax_name].default_callable(hdr)

	for k,v in axes.items():
		if len(v) == 0:
			raise RuntimeError(f'Could not work out axes "{k}" automatically, please specify axes explicitly')
	
	return FitsSpecifier(path, ext, slices, axes)



"""
Quick tool for normalising a PSF in a FITS file
"""


import sys
from pathlib import Path

from typing import Literal

import numpy as np
import scipy as sp
from astropy.io import fits

import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.specifier
import aopp_deconv_tool.astropy_helper.fits.header
import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.axes
import aopp_deconv_tool.numpy_helper.slice
import aopp_deconv_tool.arguments

import aopp_deconv_tool.psf_data_ops as psf_data_ops
from aopp_deconv_tool.fpath import FPath

import matplotlib.pyplot as plt

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')


def run(
		fits_spec : aph.fits.specifier.FitsSpecifier,
		output_path : Path,
		threshold : float = 1E-2,
		n_largest_regions : None | int = 1,
		background_threshold : float = 1E-3,
		background_noise_model : str = 'gennorm',
		n_sigma : float = 5,
		trim_to_shape : None | tuple[int,...] = None
	):
	
	original_data_type=None
	ACCEPTABLE_TRIM_LOSS_FACTOR = 1E-2
	axes = fits_spec.axes['CELESTIAL']
	
	with fits.open(Path(fits_spec.path)) as data_hdul:
		
		
		_lgr.debug(f'{fits_spec.path=} {fits_spec.ext=} {fits_spec.slices=} {fits_spec.axes=}')
		#raise RuntimeError(f'DEBUGGING')
	
		data_hdu = data_hdul[fits_spec.ext]
		data = data_hdu.data[fits_spec.slices]
		hdr = data_hdu.header
		original_data_type=data_hdu.data.dtype
		
		_lgr.info('Remove NANs and INFs')
		data[np.isnan(data) | np.isinf(data)] = 0
		
		_lgr.info('Ensure data is of odd shape')
		data = nph.array.ensure_odd_shape(data, axes)
		for ax in axes:
			aph.fits.header.set_axes_transform(hdr, ax, n_values = data.shape[ax])
		
		_lgr.info('Remove any outliers')
		outlier_mask = psf_data_ops.get_outlier_mask(data, axes, n_sigma)
		data[outlier_mask] = np.nan
		
		
		if background_noise_model is not None:
			_lgr.info('Remove background offset')
			background_mask = ~psf_data_ops.get_roi_mask(data, axes, background_threshold, 1)
			noise_model_offsets, noise_model_parameters, noise_model_at_values, noise_model_cdf, noise_model_cdf_residual = psf_data_ops.remove_offset(data, axes, background_mask, background_noise_model)
		else:
			background_mask = np.zeros_like(outlier_mask)
		
		
		
		_lgr.info('Get offsets to centre around centre of mass')
		roi_mask = psf_data_ops.get_roi_mask(data, axes, threshold, n_largest_regions)
		#center_offsets = psf_data_ops.get_centre_of_mass_offsets(data, axes, roi_mask)
		center_offsets = psf_data_ops.get_brightest_pixel_offsets(data, axes, roi_mask)
	
		_lgr.info('Recentre everything for easy comparison')
		normalised_data = psf_data_ops.apply_offsets(data, axes, center_offsets)		
		roi_mask = psf_data_ops.apply_offsets(roi_mask, axes, center_offsets)
		background_mask = psf_data_ops.apply_offsets(background_mask, axes, center_offsets)
		outlier_mask = psf_data_ops.apply_offsets(outlier_mask, axes, center_offsets)
		
		
		
		
		if trim_to_shape is not None:
			_lgr.info('Trim everything around the centre pixel')
			before_trim_sum = np.nansum(normalised_data)
			
			normalised_data = psf_data_ops.trim_around_centre(normalised_data, axes, trim_to_shape)
			roi_mask = psf_data_ops.trim_around_centre(roi_mask, axes, trim_to_shape)
			outlier_mask = psf_data_ops.trim_around_centre(outlier_mask, axes, trim_to_shape)
			
			after_trim_sum = np.nansum(normalised_data)
			trim_loss_factor = 1 - after_trim_sum/before_trim_sum
			_lgr.debug(f'{trim_loss_factor=}')
			if trim_loss_factor > ACCEPTABLE_TRIM_LOSS_FACTOR:
				_lgr.warn(f'After trimming around centre to shape {trim_to_shape}, have lost {trim_loss_factor} of original signal. This is above acceptable limit of {ACCEPTABLE_TRIM_LOSS_FACTOR}')
			
			
		
		_lgr.info('Normalise to unit sum')
		original_sum = np.nansum(normalised_data, axis=axes)
		with nph.axes.to_start(normalised_data, axes) as (gdata, gaxes):
			gdata /= original_sum
		
		if normalised_data.ndim == len(axes):
			original_sum = np.array([original_sum])
		
		
		param_dict = {
			'original_file' : Path(fits_spec.path).name, # record the file we used
			'roi_mask.threshold' : threshold,
			'roi_mask.n_largest_regions' : n_largest_regions,
			'background_mask.threshold' : background_threshold,
			'outlier_mask.n_sigma' : n_sigma,
			'trim_to_shape' : trim_to_shape,
		}
		
		hdr.update(aph.fits.header.DictReader(
			param_dict,
			prefix='psf_normalise',
			pkey_count_start=aph.fits.header.DictReader.find_max_pkey_n(hdr)
		))
		

	
	_lgr.info('save the products to a FITS file')
	hdus = []
	
	hdus.append(fits.PrimaryHDU(
		header = hdr,
		data = normalised_data.astype(original_data_type)
	))
	hdus.append(fits.ImageHDU(
		header = hdr,
		data = roi_mask.astype(int),
		name = 'ROI_MASK'
	))
	hdus.append(fits.ImageHDU(
		header = hdr,
		data = background_mask.astype(int),
		name = 'BACKGROUND_MASK'
	))
	hdus.append(fits.ImageHDU(
		header = hdr,
		data = outlier_mask.astype(int),
		name = 'OUTLIER_MASK'
	))
	hdus.append(fits.BinTableHDU.from_columns(
		columns = [
			fits.Column(name='original_total_sum', format='D', array=original_sum),
		],
		name = 'ORIG_SUM',
		header = None,
	))
	
	if background_noise_model is not None:
		hdus.append(fits.BinTableHDU.from_columns(
			columns = [
				fits.Column(name='noise_model_offsets', format='D', array=noise_model_offsets), 
			] + [
				fits.Column(name=f'noise_model_parameter_{k}', format='D', array=[x[k] for x in noise_model_parameters]) for k in noise_model_parameters[0].keys()
			],
			name = 'NOISE_MODEL_PARAMS',
			header = None,
		))
		
		hdus.append(fits.ImageHDU(
			header=None,
			name='NOISE_MODEL_CDF_VALUES',
			data = np.array(noise_model_at_values)
		))
		
		hdus.append(fits.ImageHDU(
			header=None,
			name='NOISE_MODEL_CDF',
			data = np.array(noise_model_cdf)
		))
		
		
		hdus.append(fits.ImageHDU(
			header=None,
			name='NOISE_MODEL_CDF_RESIDUAL',
			data = np.array(noise_model_cdf_residual)
		))
		
	
	hdul_output = fits.HDUList(hdus)
	hdul_output.writeto(output_path, overwrite=True)
	_lgr.info(f'Written normalised psf to {output_path.relative_to(Path().absolute(),walk_up=True)}')


def parse_args(argv):
	import os
	import aopp_deconv_tool.text
	import argparse
	
	DEFAULT_OUTPUT_TAG = '_normalised'
	DESIRED_FITS_AXES = ['CELESTIAL']
	OUTPUT_COLUMNS=80
	try:
		OUTPUT_COLUMNS = os.get_terminal_size().columns - 30
	except Exception:
		pass
	
	FITS_SPECIFIER_HELP = aopp_deconv_tool.text.wrap(
		aph.fits.specifier.get_help(DESIRED_FITS_AXES).replace('\t', '    '),
		OUTPUT_COLUMNS
	)
	
	parser = argparse.ArgumentParser(
		description=__doc__, 
		formatter_class=argparse.RawTextHelpFormatter,
		epilog=FITS_SPECIFIER_HELP
	)
	
	parser.add_argument(
		'fits_spec',
		help = '\n'.join((
			f'FITS Specifier of the data to operate upon . See the end of the help message for more information',
			f'required axes: {", ".join(DESIRED_FITS_AXES)}',
		)),
		type=str,
		metavar='FITS Specifier',
	)
	
	#parser.add_argument('-o', '--output_path', help=f'Output fits file path. By default is same as fie `fits_spec` path with "{DEFAULT_OUTPUT_TAG}" appended to the filename')
	parser.add_argument(
		'-o', 
		'--output_path', 
		type=FPath,
		metavar='str',
		default='{parent}/{stem}{tag}{suffix}',
		help = '\n    '.join((
			f'Output fits file path, supports keyword substitution using parts of `fits_spec` path where:',
			'{parent}: containing folder',
			'{stem}  : filename (not including final extension)',
			f'{{tag}}   : script specific tag, "{DEFAULT_OUTPUT_TAG}" in this case',
			'{suffix}: final extension (everything after the last ".")',
			'\b'
		))
	)
	
	parser.add_argument('--threshold', type=float, default=1E-2, help='When finding region of interest, only values larger than this fraction of the maximum value are included.')
	parser.add_argument('--n_largest_regions', type=int, default=1, help="""
		When finding region of interest, if using a threshold will only the n_largest_regions in the calculation.
		A region is defined as a contiguous area where value >= `threshold` along `axes`. I.e., in a 3D cube, if
		we recentre about the COM on the sky (CELESTIAL) axes the regions will be calculated on the sky, not in
		the spectral axis (for example)."""
	)
	parser.add_argument('--background_threshold', type=float, default=1E-3, help='Exclude the largest connected region with values larger than this fraction of the maximum value when finding the background')
	parser.add_argument('--background_noise_model', type=str, default='gennorm', 
		choices=['norm', 'gennorm', 'none'],
		help='Model of background noise to use when removing offset. "none" will mean offset is not calculated or removed'
	)	
	parser.add_argument('--n_sigma', type=float, default=5, help='When finding the outlier mask, the number of standard deviations away from the mean a pixel must be to be considered an outlier`')
	parser.add_argument('--trim_to_shape', type=int, nargs=2, default=None, help='After centreing etc. will trim data to this shape around the centre pixel. Used to reduce data volume for faster processing.')

	args = parser.parse_args(argv)
	
	args.fits_spec = aph.fits.specifier.parse(args.fits_spec, DESIRED_FITS_AXES)
	
	#if args.output_path is None:
	#	args.output_path =  (Path(args.fits_spec.path).parent / (str(Path(args.fits_spec.path).stem)+DEFAULT_OUTPUT_TAG+str(Path(args.fits_spec.path).suffix)))
	other_file_path = Path(args.fits_spec.path)
	args.output_path = args.output_path.with_fields(
		tag=DEFAULT_OUTPUT_TAG, 
		parent=other_file_path.parent, 
		stem=other_file_path.stem, 
		suffix=other_file_path.suffix
	)
	
	if args.background_noise_model == 'none':
		args.background_noise_model = None
	
	
	return args

def go(
		fits_spec, 
		output_path=None, 
		threshold=None, 
		n_largest_regions=None, 
		background_threshold=None, 
		background_noise_model=None, 
		n_sigma=None, 
		trim_to_shape=None
	):
	"""
	Thin wrapper around `run()` to accept string inputs.
	As long as the names of the arguments to this function 
	are the same as the names expected from the command line
	we can do this programatically
	"""
	# This must be first so we only grab the arguments to the function
	fargs = dict(locals().items())
	arglist = aopp_deconv_tool.arguments.construct_arglist_from_locals(fargs, n_positional_args=1)
	
	exec_with_args(arglist)
	return

def exec_with_args(argv):
	args = parse_args(argv)
	
	run(
		args.fits_spec, 
		args.output_path,
		threshold = args.threshold,
		n_largest_regions = args.n_largest_regions,
		background_threshold = args.background_threshold,
		background_noise_model = args.background_noise_model,
		n_sigma = args.n_sigma,
		trim_to_shape = args.trim_to_shape,
	)
	return
	
if __name__ == '__main__':
	exec_with_args(sys.argv[1:])


"""
Script that deconvolves the first argument with the second argument

Example invocation: 
	`python -m aopp_deconv_tool.deconvolve './example_data/test_rebin.fits{DATA}[10:12]{CELESTIAL:(1,2)}' './example_data/fit_example_psf_000.fits[10:12]{CELESTIAL:(1,2)}'`
"""

import sys, os
from pathlib import Path
from typing import Literal

from concurrent.futures import ProcessPoolExecutor as Executor, as_completed
from multiprocessing import Queue as MsgQueue
import multiprocessing as mp

import numpy as np
from astropy.io import fits

import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.specifier
import aopp_deconv_tool.astropy_helper.fits.header
import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.axes
import aopp_deconv_tool.numpy_helper.slice
import aopp_deconv_tool.psf_data_ops as psf_data_ops
from aopp_deconv_tool.fpath import FPath
import aopp_deconv_tool.arguments as arguments


from aopp_deconv_tool.algorithm.deconv.clean_modified import CleanModified
from aopp_deconv_tool.algorithm.deconv.lucy_richardson import LucyRichardson

import matplotlib as mpl
import matplotlib.cm
import matplotlib.pyplot as plt
import copy
import aopp_deconv_tool.plot_helper as plot_helper
from aopp_deconv_tool.plot_helper.base import AxisDataMapping
from aopp_deconv_tool.plot_helper.plotters import PlotSet, Histogram, VerticalLine, Image, IterativeLineGraph, IterativeLogLineGraph, HorizontalLine

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')


deconv_methods = {
	'clean_modified' : CleanModified,
	'lucy_richardson' : LucyRichardson
}

class IterationTracker:
	def __init__(self, n_iter, print_interval):
		self.i_iter = 0
		self.n_iter = n_iter
		self.print_interval = print_interval
		
	def print(self):
		if self.i_iter % self.print_interval == 0:
			self.clear()
			print(f'Iteration {self.i_iter}/{self.n_iter} [{100*self.i_iter/self.n_iter}%]', end='')
			
	def in_notebook(self):
		try:
			from IPython import get_ipython
			if 'IPKernelApp' not in get_ipython().config:  # pragma: no cover
				return False
		except ImportError:
			return False
		except AttributeError:
			return False
		return True
	
	def reset(self):
		self.i_iter=0
	
	def clear(self):
		if self.in_notebook():
			from IPython.display import clear_output
			clear_output(True)
		else:
			print('\r', end='')
	
	def update(self, *args, **kwargs):
		self.print()
		self.i_iter += 1
	
	def complete(self, *args, **kwargs):
		self.clear()
		print(f'Iteration {self.i_iter}/{self.n_iter} [{100*self.i_iter/self.n_iter}%]', end='\n')


class Iterator:
	def __init__(self, obj):
		self._iterator = iter(obj)
		self._stack = []
		self._cached = None
		self._cached_valid = False
	
	def is_valid(self):
		if not self._cached_valid:
			raise RuntimeError("Current value of iterator is invalid, maybe iterator has not started or already ended?")
	
	@property
	def next(self):
		return next(self)
	
	@property
	def current(self):
		self.is_valid()
		return self._cached
		
	def push(self, obj):
		# Push a value into the iterator at the next location
		self._stack.append(obj)
		return self
	
	def __next__(self):
		if len(self._stack) > 0:
			self._cached = self._stack.pop()
			self._cached_valid = True
			return self._cached
		
		try:
			self._cached = next(self._iterator)
			self._cached_valid = True
		except StopIteration:
			self._cached_valid = False
			raise
		else:
			return self._cached

def create_plot_set(deconvolver, cadence = 1):
	"""
	Creates a set of plots that are updated every `cadence` steps. Useful to see exactly what a deconvolver is doing.
	"""
	fig, axes = plot_helper.figure_n_subplots(8)
	axes_iter = Iterator(axes)
	a7_2 = axes[7].twinx()
	
	try:
		cmap = mpl.colormaps['bwr_oob']
	except KeyError:
		cmap = copy.copy(mpl.colormaps['bwr'])
		cmap.set_over('magenta')
		cmap.set_under('green')
		cmap.set_bad('black')
		mpl.colormaps.register(name='bwr_oob', cmap=cmap)
	
	try:
		viridis_oob = mpl.colormaps['viridis_oob']
	except KeyError:
		viridis_oob = mpl.colormaps['viridis'].copy()
		viridis_oob.set_bad(color='magenta', alpha=1)
		#viridis_oob.set_under(color='black', alpha=1)
		#viridis_oob.set_over(color='black', alpha=1)
		mpl.colormaps.register(name='viridis_oob', cmap=viridis_oob)
	
	
	
	def selected_pixels_non_selected_are_nan(x):
		r = np.array(x._selected_px)
		r[r==0] = np.nan
		return r
		
	
	
	plot_set = PlotSet(
		fig,
		'clean modified step={self.n_frames}',
		cadence=cadence,
		plots = [	
			Histogram(
				'residual', 
				static_frame=False,
				axis_data_mappings = (AxisDataMapping('value','bins',limit_getter=plot_helper.lim), AxisDataMapping('count','_hist',limit_getter=plot_helper.LimRememberExtremes()))
			).attach(axes_iter.next, deconvolver, lambda x: x._residual),
		 	
			VerticalLine(
				None, 
				static_frame=False, 
				plt_kwargs={'color':'red'}
			).attach(axes_iter.current, deconvolver, lambda x: x._pixel_threshold),
			
			Histogram(
				'residual', 
				static_frame=False,
				plt_kwargs={'color' : 'green', 'alpha':0.3},
				axis_labels=(None, 'log(count)'),
				ax_funcs=[lambda ax: ax.set_yscale('log')],
				axis_data_mappings = (AxisDataMapping('value','bins',limit_getter=plot_helper.lim), AxisDataMapping('count','_hist',limit_getter=plot_helper.LimRememberExtremes()))
			).attach(axes_iter.current.twinx(), deconvolver, lambda x: x._residual),
			
			Image(
		 		'residual'
		 	).attach(axes_iter.next, deconvolver, lambda x: x._residual),
			
			Image(
		 		'current cleaned'
			).attach(axes_iter.next, deconvolver, lambda x: x._current_cleaned),
			
			Image(
		 		'components'
			).attach(axes_iter.next, deconvolver, lambda x: x._components),
			
			Image(
		 		'selected pixels',
				plt_kwargs = {'cmap': viridis_oob},
			).attach(axes_iter.next, deconvolver, lambda x: selected_pixels_non_selected_are_nan(x)),
			
			Image(
		 		'pixel choice metric',
		 		axis_data_mappings = (AxisDataMapping('x',None), AxisDataMapping('y',None), AxisDataMapping('brightness', '_z_data', plot_helper.LimSymAroundValue(0))),
		 		plt_kwargs={'cmap':'bwr_oob'}
			).attach(axes_iter.next, deconvolver, lambda x: x._px_choice_img_ptr.val),
			
			Histogram(
				'pixel choice metric', 
				static_frame=False,
			).attach(axes_iter.next, deconvolver, lambda x: x._px_choice_img_ptr.val),
			
			Histogram(
				'pixel choice metric', 
				static_frame=False,
				plt_kwargs={'color' : 'green', 'alpha':0.3},
				axis_labels=(None, 'log(count)'),
				ax_funcs=[lambda ax: ax.set_yscale('log')]
			).attach(axes_iter.current.twinx(), deconvolver, lambda x: x._px_choice_img_ptr.val),
			
			IterativeLogLineGraph(
				'metrics',
				datasource_name='fabs',
				axis_labels = (None, 'fabs value (blue)'),
				static_frame=False,
				plt_kwargs = {},
			).attach(axes_iter.next, deconvolver, lambda x: np.fabs(np.nanmax(x._residual))),
			
			HorizontalLine(
				None, 
				static_frame=False, 
				plt_kwargs={'linestyle':'--'}
			).attach(axes_iter.current, deconvolver, lambda x: x._fabs_threshold),
			
			IterativeLogLineGraph(
				'metrics',
				datasource_name='rms',
				axis_labels = (None,'rms value (red)'),
				static_frame=False,
				plt_kwargs={'color':'red'},
			).attach(axes_iter.push(axes_iter.current.twinx()).next, deconvolver, lambda x: np.sqrt(np.nansum(x._residual**2)/x._residual.size)),
			
			HorizontalLine(
				None, 
				static_frame=False, 
				plt_kwargs={'color':'red', 'linestyle':'--'}
			).attach(axes_iter.current, deconvolver, lambda x: x._rms_threshold),
		]
	)
	return plot_set



def init_process_context(msg_queue):
	global queue
	queue = msg_queue
	

def deconv_process(group_index, idx, deconvolver, obs, psf):
	#lambda idx, obs, psf: (idx, deconvolver(obs, psf), deconvolver.get_parameters()), obs_idx, processed_obs, normed_psf))
	import sys
	from contextlib import redirect_stdout
	from time import sleep
	
	class OutFileLikeQueue:
		def __init__(self, q, group_index):
			self.q = q
			self.group_index = group_index
		
		def write(self, msg):
			#print(f'queue in: {self.group_index=} {msg=}', file=sys.stderr)
			self.q.put_nowait((self.group_index, msg))
			#sleep(0.1)
		
		def close(self):
			self.q.put((self.group_index, None))
	
	with redirect_stdout(OutFileLikeQueue(queue, group_index)) as q:
	
		try:
			result = deconvolver(obs, psf)
			params = deconvolver.get_parameters()
		except Exception:
			print(e, file=sys.stdout)
		finally:
			q.write(None)
			q.close()
		
	return group_index, idx, result, params



def display_process_progress():
	from time import sleep
	from collections import OrderedDict
	
	stop = False
	empty = True

	max_line_len = 16
	display_line = []
	most_recent_msg = dict()
	display_msg = dict()
	#print('ONE')

	while not stop:
	
		#print('TWO')
		try:
			empty = queue.empty()
		except OSError as e:
			#print('queue stopping')
			#print(e)
			stop = True
			break
		
		#print('THREE')
		if empty:
			#print('queue is empty')
			sleep(0.1)
			continue
		
		#print('FOUR')
		
		group_index, msg = queue.get()
		
		#print('FIVE')
		
		if group_index is None and msg is None:
			#print('queue stopping')
			stop = True
			break
		
		#print('SIX')
		
		if msg is None:
			#print(f'msg is "None", deleting state tracker for {group_index=}')
			if group_index in most_recent_msg:
				del most_recent_msg[group_index]
			if group_index in display_msg:
				del display_msg[group_index]
	
		
		
		#print('SEVEN')
		#print(f'queue out: {group_index=} msg="{msg}"')

		if type(msg) is str:
			most_recent_msg[group_index] = msg.replace('\r', '').strip()

		#print('EIGHT')
		if len(most_recent_msg.get(group_index, '')) == 0:
			#print('continue')
			continue
		else:
			#print('keep message')
			display_msg[group_index] = most_recent_msg.get(group_index, 'X'*16)
		
		#print('NINE')
		
		#print(' '*max_line_len, end='\r')
		
		display_line = ['- Parallel Processes ' + '-'*20] + [f'{k} {v}' for k,v in display_msg.items()]
		show_line = '\n'.join(display_line)
		
		if len(show_line) > max_line_len:
			max_line_len = len(show_line)
		
		#print(show_line, end='')
		print(show_line, end='\n')
		
		
		#print('queue out: restart loop')
	
	print(f'queue ended')
	queue.close()
		


def run(
		obs_fits_spec : aph.fits.specifier.FitsSpecifier,
		psf_fits_spec : aph.fits.specifier.FitsSpecifier,
		deconvolver : Literal[CleanModified] | Literal[LucyRichardson],
		output_path : str | Path = './deconv.fits',
		plot : bool = True,
		progress : int = 0,
		n_processes : int = (mp.cpu_count() - 2),
	):
	"""
	Given a FitsSpecifier for an observation and a PSF, an output path, and a class that performs deconvolution,
	deconvolves the observation using the PSF.
	
	# ARGUMENTS #
		obs_fits_spec : aph.fits.specifier.FitsSpecifier
			FITS file specifier for observation data, format is PATH{EXT}[SLICE](AXES).
			Where:
				PATH : str
					The path to the FITS file
				EXT : str | int
					The name or number of the FITS extension (defaults to PRIMARY)
				SLICE : "python slice format" (i.e., [1:5, 5:10:2])
					Slice of the FITS extension data to use (defaults to all data)
				AXES : tuple[int,...]
					Axes of the FITS extension that are "spatial" or "celestial" (i.e., RA, DEC),
					by default will try to infer them from the FITS extension header.
		psf_fits_spec : aph.fits.specifier.FitsSpecifier
			FITS file specifier for PSF data, format is same as above
		output_path : str = './deconv.fits'
			Path to output deconvolution to.
		deconvolver : ClassInstance
			Instance of Class to use for deconvolving, defaults to an instance of CleanModified
		plot : bool = True
			If `True` will plot the deconvolution progress
		n_processes : int = N_cpu - 2
			Number of processes to parallelise the calculations over
	"""
	
	_lgr.debug(f'{obs_fits_spec=}')
	_lgr.debug(f'{psf_fits_spec=}')
	_lgr.debug(f'{output_path=}')
	_lgr.debug(f'{deconvolver=}')
	_lgr.debug(f'{plot=}')
	_lgr.debug(f'{progress=}')
	
	original_data_type=None
	
	# Set up plotting if we want it
	if plot:
		#plt.figure()
		#plt.imshow(obs_data[obs_idx])
		#plt.figure()
		#plt.imshow(psf_data[psf_idx])
		#plt.show()
		plt.close('all')
		plot_set = create_plot_set(deconvolver, cadence=1 if progress < 1 else progress)
		deconvolver.post_iter_hooks = []
		deconvolver.post_iter_hooks.append(lambda *a, **k: plot_set.update())
		plot_set.show()

	if progress > 0:
		iteration_tracker = IterationTracker(deconvolver.n_iter, progress)
		deconvolver.post_iter_hooks.append(iteration_tracker.update)
		deconvolver.final_hooks.append(iteration_tracker.complete)
	
	
	#idx_result=[[]]
	
	# Open the fits files
	with fits.open(Path(obs_fits_spec.path)) as obs_hdul, fits.open(Path(psf_fits_spec.path)) as psf_hdul:
		
		# pull out the data we want
		obs_data = obs_hdul[obs_fits_spec.ext].data
		psf_data = psf_hdul[psf_fits_spec.ext].data
		original_data_type=obs_data.dtype
		
		group_shape = tuple(s for x, s in enumerate(obs_data.shape) if x not in obs_fits_spec.axes['CELESTIAL'])
		idx_result = np.full(group_shape, dtype=object, fill_value=dict())
		
		_lgr.debug(f'{obs_data.shape=}')
		_lgr.debug(f'{psf_data.shape=}')
		_lgr.debug(f'{original_data_type=}')
		
		# Create holders for deconvolution products
		deconv_components = np.full_like(obs_data, np.nan)
		deconv_residual = np.full_like(obs_data, np.nan)
		
		msg_queue = MsgQueue()
		
		with Executor(max_workers=n_processes, initializer=init_process_context, initargs=(msg_queue,)) as executor:
			
			msg_responder_process = executor.submit(display_process_progress)
			
			futures = []
			# Loop over the index range specified by `obs_fits_spec` and `psf_fits_spec`
			for i, (obs_idx, psf_idx) in enumerate(zip(
					nph.slice.iter_indices(obs_data, obs_fits_spec.slices, obs_fits_spec.axes['CELESTIAL']),
					nph.slice.iter_indices(psf_data, psf_fits_spec.slices, psf_fits_spec.axes['CELESTIAL'])
				)):
				if progress > 0:
					iteration_tracker.reset()
					print(f'Deconvolving slice {i}/{obs_data[obs_fits_spec.slices].shape[tuple(x for x in range(obs_data.ndim) if x not in obs_fits_spec.axes['CELESTIAL'])[0]]}')
				_lgr.debug(f'Operating on slice {i}')
				
				
				# Get group index we can use to uniquely identify which
				# section of the array we are operating on
				gi_elems = tuple(0 for x in range(obs_data.ndim) if x not in obs_fits_spec.axes['CELESTIAL'])
				group_index = []
				group_obs_idx = obs_idx
				counter = 0
				for ax in sorted(obs_fits_spec.axes['CELESTIAL'], reverse=True):
					group_obs_idx = np.take(group_obs_idx, 0, ax-counter)
					counter += 1
				for _i,_v in enumerate(gi_elems):
					group_index.append(np.take(group_obs_idx, _i, axis=_v))
				group_index = tuple(group_index)
				
				
				# Ensure that we actually have data in this part of the cube
				if np.all(np.isnan(obs_data[obs_idx])) or np.all(np.isnan(psf_data[psf_idx])):
					_lgr.warn('All NAN obs or psf layer detected. Skipping...')
					continue
				
				# perform any normalisation and processing
				normed_psf = psf_data_ops.normalise(np.nan_to_num(psf_data[psf_idx]))
				processed_obs = np.nan_to_num(obs_data[obs_idx])
				
				"""
				# Store the deconvolution products in the arrays we created earlier
				deconv_components[obs_idx], deconv_residual[obs_idx], deconv_iters = deconvolver(processed_obs, normed_psf)
				
				# Save the parameters we used. NOTE: we are only saving the LAST set of parameters as they are all the same
				# in this case. However, if they vary with index they should be recorded with index as well.
				deconv_params = deconvolver.get_parameters()
			
				idx_result.append({**deconv_params})
				"""
				
				futures.append(executor.submit(deconv_process, group_index, obs_idx, deconvolver, processed_obs, normed_psf))
			
			print(f'Waiting for processes...')
			for future in as_completed(futures):
				if not future.done():
					_lgr.error(f'Deconv process {future} not completed, something went wrong')
					continue
				if future.cancelled():
					_lgr.warn(f'Deconv process {fugure} cancelled, skipping result...')
					continue
				
				_lgr.info(f'Process {group_index} ended')
				msg_queue.put((group_index, None))
				
				group_index, idx, (future_deconv_components, future_deconv_residual, future_deconv_iters), future_params = future.result()
				
				_lgr.debug(f'{group_index=}')
				_lgr.debug(f'{idx=}')
				_lgr.debug(f'{future_deconv_components=}')
				_lgr.debug(f'{future_deconv_residual=}')
				_lgr.debug(f'{future_deconv_iters=}')
				_lgr.debug(f'{future_params=}')
				deconv_components[idx] = future_deconv_components
				deconv_residual[idx] = future_deconv_residual
				
				idx_result[*group_index] = {**future_params}

			_lgr.info(f'Processes completed')
			msg_queue.put((None,None))
			msg_queue.close()
	

	_lgr.debug(f'{idx_result=}')
	
	param_names = None
	param_defaults = None
	for item in idx_result:
		if len(item) > 0:
			param_names = tuple(item.keys())
			param_defaults = item
			break
	
	params_that_change = []
	for k in param_names:
		if any((x[k] != param_defaults[k] for x in idx_result if len(x) != 0)):
			params_that_change.append(k)
	params_that_are_static=[k for k in param_names if k not in params_that_change]
	
	vp_type_missing = {}
	vp_type_codes = {}
	for k in params_that_change:
		v = param_defaults[k]
		if type(v) is str:
			vp_type_codes[k] = 'A'
			vp_type_missing[k] = 'MISSING'
		elif type(v) is int:
			vp_type_codes[k]='I'
			vp_type_missing[k] = -1
		elif type(v) is float:
			vp_type_codes[k]='D'
			vp_type_missing[k] = np.nan
		else:
			vp_type_codes[k]='1024A'
			vp_type_missing[k] = 'MISSING'
	
	static_params = dict(((k,param_defaults[k]) for k in param_names if k in params_that_are_static))
	volatile_params = dict((k, [x.get(k, vp_type_missing[k]) for x in idx_result]) for k in params_that_change)
	
	
	
	for k,v in vp_type_codes.items():
		if v[0]=='A':
			vp_type_codes[k]=f'{max(map(len,volatile_params[k]))}A'
	
	
	# Make sure we get all the observaiton header data as well as the deconvolution parameters
	hdr = obs_hdul[obs_fits_spec.ext].header
	hdr.update(aph.fits.header.DictReader(
		{
			'obs_file' : obs_fits_spec.path,
			'psf_file' : psf_fits_spec.path, # record the PSF file we used
			**static_params # record the deconvolution parameters we used
		},
		prefix='deconv',
		pkey_count_start=aph.fits.header.DictReader.find_max_pkey_n(hdr)
	))
	
	# Save the deconvolution products to a FITS file
	hdu_components = fits.PrimaryHDU(
		header = hdr,
		data = deconv_components.astype(original_data_type)
	)
	hdu_residual = fits.ImageHDU(
		header = hdr,
		data = deconv_residual.astype(original_data_type),
		name = 'RESIDUAL'
	)
	hdu_deconv_param_tbl = fits.BinTableHDU.from_columns(
		columns = [fits.Column(name=f'deconv.{k}', format=vp_type_codes[k], array=v) for k,v in volatile_params.items()],
		name = 'VOLATILE_PARAMS',
		header = None,
	)
	
	
	hdul_output = fits.HDUList([
		hdu_components,
		hdu_residual,
		hdu_deconv_param_tbl
	])
	hdul_output.writeto(output_path, overwrite=True)
	
	_lgr.info(f'Deconvolution completed, output written to "{output_path}"')
	

def parse_args(argv):
	import os
	import aopp_deconv_tool.text
	import argparse
	
	DEFAULT_OUTPUT_TAG = '_deconv'
	DESIRED_FITS_AXES = ['CELESTIAL']
	OUTPUT_COLUMNS=80
	try:
		OUTPUT_COLUMNS = os.get_terminal_size().columns - 30
	except Exception:
		pass
	
	FITS_SPECIFIER_HELP = aopp_deconv_tool.text.wrap(
		aph.fits.specifier.get_help(DESIRED_FITS_AXES).replace('\t', '    '),
		OUTPUT_COLUMNS
	)
	DECONV_METHOD_DEFAULT='clean_modified'
	
	class ArgFormatter (argparse.RawTextHelpFormatter, argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter):
		def __init__(self, *args, **kwargs):
			super().__init__(*args, **kwargs)
	
	
	
	
	
	parser = argparse.ArgumentParser(
		description=__doc__, 
		formatter_class=ArgFormatter,
		epilog=FITS_SPECIFIER_HELP,
		exit_on_error=False
	)
	
	parser.add_argument(
		'obs_fits_spec',
		help = '\n'.join((
			f'The observation\'s (i.e., science target) FITS Specifier. See the end of the help message for more information',
			f'required axes: {", ".join(DESIRED_FITS_AXES)}',
		)),
		type=str,
		metavar='FITS Specifier',
	)
	
	parser.add_argument(
		'psf_fits_spec',
		help = '\n'.join((
			f'The psf\'s (i.e., calibration target) FITS Specifier. See the end of the help message for more information',
			f'required axes: {", ".join(DESIRED_FITS_AXES)}',
		)),
		type=str,
		metavar='FITS Specifier',
	)
	
	parser.add_argument(
		'-o', 
		'--output_path', 
		type=FPath,
		metavar='str',
		default='{parent}/{stem}{tag}{suffix}',
		help = '\n    '.join((
			f'Output fits file path, supports keyword substitution using parts of `obs_fits_spec` path where:',
			'{parent}: containing folder',
			'{stem}  : filename (not including final extension)',
			f'{{tag}}   : script specific tag, "{DEFAULT_OUTPUT_TAG}" in this case',
			'{suffix}: final extension (everything after the last ".")',
			'\b'
		))
	)
	parser.add_argument(
		'--plot', 
		action='store_true', 
		default=False, 
		help='If present will show progress plots of the deconvolution'
	)
	parser.add_argument(
		'--deconv_method', 
		type=str, 
		choices=deconv_methods.keys(), 
		default=DECONV_METHOD_DEFAULT, 
		help='Which method to use for deconvolution. For more information, pass the deconvolution method and the "--info" argument.'
	) 
	parser.add_argument(
		'--deconv_method_help', 
		action='store_true', 
		default=False, 
		help='Show help for the selected deconvolution method'
	)
	parser.add_argument(
		'--progress',
		type=int,
		default=0,
		help='Show progression of deconvolution on each `progress` step, 0 does not display progress'
	)
	
	parser.add_argument(
		'--n_processes',
		type=int,
		default=mp.cpu_count() - 2,
		help= 'Number of processors to parallelise the calculations over'
	)
	
	parser.successful = True
	parser.error_message = None
	
	def on_error(err_str):
		parser.successful = False
		parser.error_message = err_str
		
	parser.error = on_error
	
	args, deconv_args = parser.parse_known_args(argv)
	
	if not parser.successful:
		print(vars(args))
		if args.deconv_method_help:
			return args, deconv_args
		else:
			parser.print_usage()
			print(parser_error_message)
			sys.exit()
			
	
	
	args.obs_fits_spec = aph.fits.specifier.parse(args.obs_fits_spec, DESIRED_FITS_AXES)
	args.psf_fits_spec = aph.fits.specifier.parse(args.psf_fits_spec, DESIRED_FITS_AXES)
	
	other_file_path = Path(args.obs_fits_spec.path)
	args.output_path = args.output_path.with_fields(
		tag=DEFAULT_OUTPUT_TAG, 
		parent=other_file_path.parent, 
		stem=other_file_path.stem, 
		suffix=other_file_path.suffix
	)
	
	return args, deconv_args


def go(
		obs_fits_spec,
		psf_fits_spec,
		output_path=None, 
		plot=None, 
		deconv_method=None, 
		deconv_method_help_FLAG=None,
		n_processes=None,
		**kwargs
	):
	"""
	Thin wrapper around `run()` to accept string inputs.
	As long as the names of the arguments to this function 
	are the same as the names expected from the command line
	we can do this programatically
	"""
	# This must be first so we only grab the arguments to the function
	fargs = dict(locals().items())
	arglist = aopp_deconv_tool.arguments.construct_arglist_from_locals(fargs, n_positional_args=2)
	
	exec_with_args(arglist)
	return

def exec_with_args(argv):
	args, deconv_args = parse_args(argv)
	_lgr.debug('#### ARGUMENTS ####')
	for k,v in vars(args).items():
		_lgr.debug(f'\t{k} : {v}')
	_lgr.debug('#### END ARGUMENTS ####')
	
		
	deconv_class = deconv_methods[args.deconv_method]
	deconv_params = arguments.parse_args_of_dataclass(
		deconv_class, 
		deconv_args, 
		prog=f'deconvolve.py --deconv_method {args.deconv_method}',
		show_help=args.deconv_method_help
	)
	
	_lgr.debug(f'{deconv_params=}')
	deconvolver = deconv_class(**deconv_params)
	
	run(
		args.obs_fits_spec, 
		args.psf_fits_spec, 
		deconvolver = deconvolver,
		output_path = args.output_path, 
		plot = args.plot,
		progress = args.progress,
		n_processes = args.n_processes,
	)
	
	return
	
if __name__ == '__main__':
	argv = sys.argv[1:]
	exec_with_args(argv)
	



"""
Padding operations
"""

import scipy as sp
import numpy as np

import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.slice

def with_convolution(a : np.ndarray, b : np.ndarray) -> np.ndarray:
	"""
	Pad array `a` by convolving it with array `b`, using the edges of array `b`
	and the centre is array `a`
	"""
	p = sp.signal.fftconvolve(a, b, mode='full')
	p[nph.slice.around_centre(p.shape,a.shape)] = a
	return(p)

def with_const(a : np.ndarray, to_shape : tuple[int,...], const = 0) -> np.ndarray:
	"""
	Padd array `a` to shape `to_shape` with the value `const`
	"""
	p = np.full(to_shape, const, dtype=a.dtype)
	p[nph.slice.around_centre(to_shape, a.shape)] = a
	return p




from typing import Literal

import numpy as np
import scipy as sp
import scipy.stats
import scipy.ndimage
import matplotlib.pyplot as plt

import aopp_deconv_tool.context as context
from aopp_deconv_tool.context.next import Next

from aopp_deconv_tool.stats.empirical import EmpiricalDistribution
from aopp_deconv_tool.scipy_helper.interp import interpolate_at_mask

import aopp_deconv_tool.plot_helper as plot_helper


import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')


def ssa2d_deviations(
		ssa,
		start = None,
		stop = None,
		size = None,
		mask = None,
	) -> np.ndarray:
	"""
	Average standard deviation of a pixel from a region `size` around it.
	Same as `ssa2d_sum_prob_map` but a bit more efficient (less customisable).
	"""
	
	size = max(ssa.a.shape)//4 if size is None else size
	mask = np.ones(ssa.a.shape, dtype=bool) if mask is None else mask
	
	deviation_sum = np.zeros(ssa.a.shape, dtype=float)
	ssa_means = np.zeros(ssa.a.shape, dtype=float)
	ssa_std = np.zeros(ssa.a.shape, dtype=float)
	for i in range(stop-start):
		#_lgr.debug(f'{i=}')
		#deviation_sum += sp.ndimage.generic_filter(ssa.X_ssa[start+i]-ssa_means[i], kernel, size=10)
		x = ssa.X_ssa[start+i]

		ssa_means = sp.ndimage.uniform_filter(x, size=size)
		#_lgr.debug(f'{ssa_means.shape=}')
		ssa_std = np.sqrt(sp.ndimage.uniform_filter((x - ssa_means)**2, size=size))
		#_lgr.debug(f'{ssa_std.shape=}')
		deviation_sum[mask] += ((x - ssa_means)/(ssa_std + 1E-2*np.nanmean(ssa_std)))[mask] # number of standard deviations away from mean
		
	return np.fabs(deviation_sum)/(stop-start)


def ssa2d_sum_prob_map(
		ssa, 
		start=5, 
		stop=None, 
		value=0.99, 
		strategy : Literal['prob_frac_from_median'] | Literal['n_std_dev_from_median'] = 'n_std_dev_from_median',
		show_plots=0, 
		transform_value_as : list[Literal['median_prob', 'ppf'],] = ['ppf'], 
		weight_by_evals=False,
		smooth_sigma=None,
	):
	"""
	Computes a bad pixel map from an SSA2D object. Concentrates on hot/cold pixels w.r.t background.
	
	Idea is that for a hot/cold pixel, the high frequency components of the SSA decomposition must 
	also be hot/cold. Therefore, the hot/cold pixels will consistently be at the *edges* of the
	pixel probability distribution for the higher frequency SSA components. As high frequency
	SSA components outnumber low frequency SSA components, we can just use all components as 
	long as we don't weight them by their eigenvalues (if we do, we should try and be selective
	to only the components where the scale of changes matches the scale of the hot/cold pixels).
	
	For each component of the SSA decomposition, a map is made of how far a given pixel deviates 
	from the median (range [-1,1]). The median differences for each component are combined together
	(by mean or weighted mean) to give a 'score' for each pixel. Bad pixels should have scores at the
	extreme ends of the [-1,1] range. A bad pixel map is chosen by taking the absolute value of the score,
	and selecting all pixels whose |score| is larger than a give value.

	# ARGUMENTS #
		ssa
			An SSA2D object
		start
			The first SSA component to include in the probability map
		stop
			The last SSA component to include in the probability map.
			If None, then will include as many SSA components as are present
		value
			A float that describes the quantile that denotes a 'bad pixel'
		strategy : 'prob_frac_from_median' | 'n_std_dev_from_median'
			How should we define an 'improbable' pixel?
				'prob_frac_from_median'
					Fraction of probability space a pixel is away from 
					the median value
				'n_std_dev_from_median'
					Number of standard deviations a pixel is away from
					the median value
		show_plots
			0 = do not show plots
			1 = show some plots
			2 = show interim plots
		transform_value_as
			List of literal strings that tells us how to tranform the "value"
			argument after which it is used as the cutoff for a bad pixel.
			If no options are given, "value" is used as the cutoff directly,
			i.e., if "value" is 0.99, then only pixels which have a median
			probability > 0.99 are classified as "bad pixels".

			Options are:
				"identity"
					Use the value as-is, do not transform it
				"median_prob"
					Transform "value" using the same function that we use to
					change cumulative probabilities of pixels into probability
					a pixel is a certain distance from the median.
				"ppf"
					Transform "value" so that the cutoff is the percentage
					point function of the pixel median probabilities. I.e.
					if "value" is 0.9, then the cutoff will be set such that
					10% of the pixels will be classified as "bad pixels"
		weight_by_evals
			If True, will weight each SSA component pixel probability by
			the component's eigenvalue when averaging into a combined score.
	# RETURNS #
		bp_mask
			A bad pixel mask. Pixels to reject are coded as True.
	"""
	# set up arrays and defaults
	stop = ssa.X_ssa.shape[0] if stop is None else stop
	start = 0 if start is None else start
	distrib = sp.stats.cauchy
	data_probs = np.zeros((stop-start, *ssa.a.shape),dtype=float)
	
	# choose a pixel->probability function
	# we get a range of (0,1) from "ut.sp.construct_cdf_from()"
	# it gives us the cumulative probability for each pixel
	# This function changes it to 'distance away from median' in the range [-1, 1]
	prob_median_transform_func = lambda x: (2*(x-0.5))
	
	# apply pixel->probability function to each SSA component
	for i in range(0, stop-start):
		match strategy:
			case 'prob_frac_from_median':
				data_distribution = EmpiricalDistribution(ssa.X_ssa[i+start].ravel())
				data_probs[i,...] = prob_median_transform_func(data_distribution.cdf(ssa.X_ssa[i+start].ravel()).reshape(ssa.a.shape))
			case 'n_std_dev_from_median':
				#data_probs[i,...] = (ssa.X_ssa[i+start] - np.nanmean(ssa.X_ssa[i+start]))/np.nanstd(ssa.X_ssa[i+start])
				data_probs[i,...] = (ssa.X_ssa[i+start] - np.nanmedian(ssa.X_ssa[i+start]))/np.nanstd(ssa.X_ssa[i+start])
				#data_probs[i,...] = np.fabs((ssa.X_ssa[i+start] - np.nanmedian(ssa.X_ssa[i+start]))/np.nanstd(ssa.X_ssa[i+start]))
			case 'magnitude':
				data_probs[i,...] = np.fabs(ssa.X_ssa[i+start])
			case 'identity':
				data_probs[i,...] = ssa.X_ssa[i+start]
			case _:
				raise RuntimeError(f'Unknown `strategy` for getting probability of a pixel {strategy=}')

	# combine pixel probabilites together, pixels that are "strange" will have correlated distances away from
	# the median for each SSA component. Therefore, "normal" pixels will tend towards zero when taking the mean of `data_probs`,
	# but "strange" (hot/cold) pixels will stay close to +/- 1. Finally, then take absolute value 
	# as we want a range of [0,1] where 0 is data is "normal" and 1 is data is "strange"
	# NOTE: Instead of using `np.fabs` here, I could square the results. It shouldn't make much of a difference apart from adjusting
	#       the exact values of the `data_probs_sum`.
	data_probs_sum_func = lambda x, _start, _stop: np.fabs(
		np.mean(x[:_stop-(_start-1)], axis=0)
	)
	if weight_by_evals:
		# try weighted by eigenvalues, generally do not want this one but the option is there
		data_probs_sum_func = lambda x, _start, _stop: np.fabs(
			(1/(np.sum((np.diag(ssa.s)**2)[_start:_stop]))) * np.sum(((np.diag(ssa.s)**2)[_start:_stop])[:,None,None] * x[:_stop-(_start-1)],axis=0)
		)
	
	
	if smooth_sigma is not None:
		old_data_probs_sum_func = data_probs_sum_func
		data_probs_sum_func = lambda x, _start, _stop: np.fabs(
			sp.ndimage.gaussian_filter(old_data_probs_sum_func(x, _start, _stop), smooth_sigma, order=1)
		)
	

	
	data_probs_sum = data_probs_sum_func(data_probs,start,stop)
	

	# plots for debugging and progress
	if show_plots > 1:
		f1, a1 = plot_helper.figure_n_subplots(9)
		f1.suptitle(f'Sum ssa.X_ssa[{start}:{stop}] probability maps')
		ax_iter=iter(a1.flatten())
		
		i = start
		with Next(ax_iter) as ax:
			ax.set_title(f'ssa.X_ssa[{i}]\nsum {np.nansum(ssa.X_ssa[i])}')
			ax.imshow(ssa.X_ssa[i])
			plot_helper.remove_axes_ticks_and_labels(ax)
		with Next(ax_iter) as ax:
			ax.set_title(f'|probabilities of ssa.X_ssa[{i}]|')
			#ax.imshow(data_probs_sum_func(data_probs, start, i), vmin=0, vmax=1)
			ax.imshow(data_probs[i-start])
			plot_helper.remove_axes_ticks_and_labels(ax)
		with Next(ax_iter) as ax:
			ax.set_title(f'histogram of ssa.X_ssa[{i}] range [{np.min(ssa.X_ssa[i])}, {np.max(ssa.X_ssa[i])}] sigma {np.std(ssa.X_ssa[i])}')
			hvals, hbins, hpatches = ax.hist(ssa.X_ssa[i].ravel(), bins=100, density=True)
			x = np.linspace(np.min(ssa.X_ssa[i]),np.max(ssa.X_ssa[i]), 100)
			ax.twinx().plot(x, EmpiricalDistribution(ssa.X_ssa[i].ravel()).cdf(x), color='tab:orange', ls='-')
			ax.set_yscale('log')
		
		i = stop -1
		with Next(ax_iter) as ax:
			ax.set_title(f'ssa.X_ssa[{i}]\nsum {np.nansum(ssa.X_ssa[i])}')
			ax.imshow(ssa.X_ssa[i])
			plot_helper.remove_axes_ticks_and_labels(ax)
		with Next(ax_iter) as ax:
			ax.set_title(f'|probabilities of ssa.X_ssa[{i}]|')
			#ax.imshow(data_probs_sum_func(data_probs, start, i), vmin=0, vmax=1)
			ax.imshow(data_probs[i-start])
			plot_helper.remove_axes_ticks_and_labels(ax)
		with Next(ax_iter) as ax:
			ax.set_title(f'histogram of ssa.X_ssa[{i}] range [{np.min(ssa.X_ssa[i])}, {np.max(ssa.X_ssa[i])}] sigma {np.std(ssa.X_ssa[i])}')
			hvals, hbins, hpatches = ax.hist(ssa.X_ssa[i].ravel(), bins=100, density=True)
			x = np.linspace(np.min(ssa.X_ssa[i]),np.max(ssa.X_ssa[i]), 100)
			ax.twinx().plot(x, EmpiricalDistribution(ssa.X_ssa[i].ravel()).cdf(x), color='tab:orange', ls='-')
			ax.set_yscale('log')
		
		i = stop
		with Next(ax_iter) as ax:
			ax.set_title(f'|sum of probabilities of ssa.X_ssa[{start}:{i}]|')
			#ax.imshow(data_probs_sum_func(data_probs, start, i), vmin=0, vmax=1)
			ax.imshow(data_probs_sum_func(data_probs, start, i))
			plot_helper.remove_axes_ticks_and_labels(ax)
		with Next(ax_iter) as ax:
			ax.set_title(f'histogram of sum of probabilities of ssa.X_ssa[{start}:{i}]|')
			z = data_probs_sum_func(data_probs, start, i)
			hvals, hbins, hpatches = ax.hist(z.ravel(), bins=100, density=True)
			x = np.linspace(np.min(z),np.max(z), 100)
			ax.twinx().plot(x, EmpiricalDistribution(z.ravel()).cdf(x), color='tab:orange', ls='-')
			ax.set_yscale('log')
		with Next(ax_iter) as ax:
			ax.set_title(f'log(|sum of probabilities of ssa.X_ssa[{start}:{i}]|)')
			#ax.imshow(data_probs_sum_func(data_probs, start, i), vmin=0, vmax=1)
			ax.imshow(np.log(data_probs_sum_func(data_probs, start, i)))
			plot_helper.remove_axes_ticks_and_labels(ax)
			
		plot_helper.output(
			True, 
			None,
			#f'ssa_{i}_bad_pixel_prob_maps.png',
		)
	# plots for debugging and progress
	if show_plots > 2:
		nplots = ssa.X_ssa.shape[0]#(stop-start)
		
		#for i in range(nplots):
		#for i in range(start, stop, (stop-start)//5):
		for i in [stop-1]:
			f1, a1 = plot_helper.figure_n_subplots(4)
			f1.suptitle(f'Sum ssa.X_ssa[{i}/{nplots}] probability maps')
			ax_iter=iter(a1.flatten())
			
			with Next(ax_iter) as ax:
				ax.set_title(f'ssa.X_ssa[{i}]\nsum {np.nansum(ssa.X_ssa[i])}')
				ax.imshow(ssa.X_ssa[i])
				plot_helper.remove_axes_ticks_and_labels(ax)
			with Next(ax_iter) as ax:
				ax.set_title(f'histogram of ssa.X_ssa[{i}] range [{np.min(ssa.X_ssa[i])}, {np.max(ssa.X_ssa[i])}] sigma {np.std(ssa.X_ssa[i])}')
				hvals, hbins, hpatches = ax.hist(ssa.X_ssa[i].ravel(), bins=100, density=True)
				x = np.linspace(np.min(ssa.X_ssa[i]),np.max(ssa.X_ssa[i]), 100)
				ax.twinx().plot(x, EmpiricalDistribution(ssa.X_ssa[i].ravel()).cdf(x), color='tab:orange', ls='-')
			if i < stop and i >= start:
				with Next(ax_iter) as ax:
					ax.set_title(f'probabilities of ssa.X_ssa[{i}]')
					#ax.imshow(data_probs[i-start], vmin=-1, vmax=1)
					ax.imshow(data_probs[i-start])
					plot_helper.remove_axes_ticks_and_labels(ax)
				with Next(ax_iter) as ax:
					ax.set_title(f'|sum of probabilities of ssa.X_ssa[{start}:{i+1}]|')
					#ax.imshow(data_probs_sum_func(data_probs, start, i), vmin=0, vmax=1)
					ax.imshow(data_probs_sum_func(data_probs, start, i))
					plot_helper.remove_axes_ticks_and_labels(ax)
				
			plot_helper.output(
				True, 
				None,
				#f'ssa_{i}_bad_pixel_prob_maps.png',
			)
	
	return data_probs_sum
	
def get_bp_mask_from_badness_map(badness_map, value, value_transform = 'identity'):
	# apply the pixel->probability function to 'value' argument if desired
	"""
	if 'median_prob' in transform_value_as:
		value = prob_median_transform_func(value)
	if 'ppf' in transform_value_as:
		cutoff_func = lambda _test, _value: EmpiricalDistribution(_test.ravel()).ppf(_value)
	else:
		cutoff_func = lambda _test, _value: _value
	"""
	
	match value_transform:
		case 'ppf':
			cutoff_func = lambda _test, _value: EmpiricalDistribution(_test.ravel()).ppf(_value)
		case 'identity':
			cutoff_func = lambda _test, _value: _value
		case _:
			raise RuntimeError(f'Unknown value transform "{value_transform}"')
	
	
	
	bp_mask = pixel_map(
		badness_map, 
		value, 
		show_plots=False,
		img=None,
		cutoff_func=cutoff_func, 
		plot_kw = {
			'suptitle':f'Sum ssa2d probability maps\n{value=}'
		}
	)
		
	return(bp_mask)



def pixel_map(
		test, 
		value, 
		show_plots=0, 
		img = None,
		plot_kw={}, 
		cutoff_func = lambda _test, _value: EmpiricalDistribution(_test.ravel()).ppf(_value), 
		bp_mask_func = lambda _test, _cutoff: _test > _cutoff
	) -> np.ndarray:
	"""
	Finds a mask for an image based on values in 'test'.

	# ARGUMENTS #
		img : np.ndarray([:,:], type.T1)
			image to construct mask for (not actually used in calculation, but used in plotting)
		test : np.ndarray([:,:], type.T2)
			Array that cutoff will be applied to to find mask
		value : type.T3
			Value to calculate cutoff from using 'cutoff_func'
		show_plots : bool
			If True, will plot results of mapping.
		plot_kw : dict
			Dictionary to pass optional arguments to plots.
			{	'suptitle' : str
					A string to display as the super-title of the plotted figure
			}
		cutoff_func : Callable[np.ndarray([:,:],type.T2), type.T3] -> type.T4
				= lambda _t, _v: ut.sp.construct_ppf_from(_t.ravel())(_v)
			A function that calculates the cutoff from 'value'. The cutoff is used
			to detemine the mask.
		bp_mask_func : Callable[np.ndarray([:,:],type.T2), type.T4] -> np.ndarray([:,:],bool)
				= lambda _t, _c: _t > _c
			A function that calculates the mask from the test and cutoff (cutoff is calculated
			from 'cutoff_func')
			
	# RETURNS #
		mask : np.ndarray([:,:],bool)
			A boolean array that can be used to mask 'img'
	"""
	cutoff = cutoff_func(test, value)
	bp_mask = bp_mask_func(test, cutoff)

	#breakpoint() # DEBUGGING
	if show_plots > 0 and img is not None:
		interpolated = interpolate_at_mask(img, bp_mask, edges='convolution', method='cubic')
		plot_pixel_map_test(img, test, cutoff, bp_mask, interpolated, plot_kw=plot_kw)
		plot_helper.output(
			True,
			None,
			#f'pixel_map_plot.png'
		)
	return(bp_mask)


def plot_pixel_map_test(img, test, cutoff, mask, interp, plot_kw={}):
	f2, a2 = plot_helper.figure_n_subplots(6)
	a2_iter = iter(a2.ravel())
	f2.suptitle(plot_kw.get('suptitle', 'pixel map function'))
	
	with Next(a2_iter) as ax:
		ax.set_title(f'Original image\nsum {np.nansum(img):08.2E}\nsqrt(sum^2) {np.sqrt(np.nansum(img**2)):08.2E}')
		ax.imshow(img)
		plot_helper.remove_axes_ticks_and_labels(ax)
		
	with Next(a2_iter) as ax:
		ax.set_title(f'pixel choice test function\nsum {np.nansum(test):08.2E} sqrt(sum^2) {np.sqrt(np.nansum(test**2)):08.2E}')
		#ax.imshow(test)
		ax.imshow(np.log(test))
		plot_helper.remove_axes_ticks_and_labels(ax)
	
	with Next(a2_iter) as ax:
		ax.set_title(f'histogram of choice function\ncutoff {cutoff:06.4f}')
		ax.hist(test.ravel(), bins=100)
		ax.axvline(cutoff, color='red', ls='--')
		ax.set_yscale('log')
	
	with Next(a2_iter) as ax:
		ax.set_title(f'mask from cut of choice function\nn_masked {np.nansum(mask)} frac_masked {np.nansum(mask)/mask.size:08.2E}')
		ax.imshow(mask)
		plot_helper.remove_axes_ticks_and_labels(ax)
		
	with Next(a2_iter) as ax:
		ax.set_title('\n'.join((f'original interpolated at mask',
								f'sum {np.nansum(interp):08.2E} frac {np.nansum(interp)/np.nansum(img):08.2E}',
								f'sqrt(sum^2) {np.sqrt(np.nansum(interp**2)):08.2E} frac {np.sqrt(np.nansum(interp**2))/np.sqrt(np.nansum(img**2)):08.2E}',
		)))
		ax.imshow(interp)
		plot_helper.remove_axes_ticks_and_labels(ax)
		
	with Next(a2_iter) as ax:
		residual = interp - img
		ax.set_title('\n'.join((f'residual of interpolation - original',
								f'sum {np.nansum(residual):08.2E} frac {np.nansum(residual)/np.nansum(img):08.2E}',
								f'sqrt(sum^2) {np.sqrt(np.nansum(residual**2)):08.2E} frac {np.sqrt(np.nansum(residual**2))/np.sqrt(np.nansum(img**2)):08.2E}',
		)))
		ax.imshow(residual)
		plot_helper.remove_axes_ticks_and_labels(ax)



"""
Quick tool for spatially rebinning a FITS file, useful for reducing data volume when testing.

Linearly interpolates at a new grid starting at the same location as the old one, but with a different step size. Uses the same units the fits file is in.

"""
import sys, os
import numpy as np
import numpy.linalg
import scipy as sp
import scipy.interpolate
import scipy.ndimage

from pathlib import Path

import matplotlib.pyplot as plt
from astropy.io import fits

from aopp_deconv_tool.fpath import FPath
import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.specifier
import aopp_deconv_tool.astropy_helper.fits.header
import aopp_deconv_tool.arguments

import aopp_deconv_tool.numpy_helper as nph

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')




def run(
		fits_spec : aph.fits.specifier.FitsSpecifier,
		output_path : str | Path,
		steps : tuple[float,float],
	):
	with fits.open(Path(fits_spec.path)) as data_hdul:
	
		data_hdu = data_hdul[fits_spec.ext]
		data = data_hdu.data
		hdr = data_hdu.header
		axes_ordering =  aph.fits.header.get_axes_ordering(data_hdu.header, fits_spec.axes['CELESTIAL'])
		ax_values = aph.fits.header.get_world_coords_of_axis(data_hdu.header, fits_spec.axes['CELESTIAL'])
		iwc_matrix = aph.fits.header.get_iwc_matrix(hdr)
		
		_lgr.debug(f'{iwc_matrix=}')
		
		transform_matrix = np.array([
			[1,                                0,                                0],
			[0, np.abs(steps[0]/iwc_matrix[1,1]),                                0],
			[0,                                0, np.abs(steps[1]/iwc_matrix[2,2])],
		])
		_lgr.debug(f'{transform_matrix=}')
		
		result = np.full(tuple(int(np.floor(data.shape[i]/transform_matrix[i,i])) for i in range(data.ndim)), fill_value=np.nan)
		_lgr.debug(f'{result.shape=}')
		
		for i, (idx, group_slice) in enumerate(nph.slice.iter_indices_with_slices(data, fits_spec.slices, fits_spec.axes['CELESTIAL'])):
			_lgr.info(f'Operating on slice {i}')
			sp.ndimage.affine_transform(
				np.nan_to_num(data[idx]), 
				transform_matrix[fits_spec.axes['CELESTIAL'],fits_spec.axes['CELESTIAL']], 
				order=1, # linear interpolation
				output_shape=tuple(s for i,s in enumerate(result.shape) if i in fits_spec.axes['CELESTIAL']), 
				output=result[group_slice],
				mode='constant',
				cval=np.nan
			)

		axis_fits = tuple(ao.fits for ao in axes_ordering)
		param_dict = {
			'original_file' : Path(fits_spec.path).name, # record the file we used
			'axes' : axis_fits,
			'rebinned_steps' : steps,
		}
		
		hdr.update(aph.fits.header.DictReader(
			param_dict,
			prefix='spatial_rebin',
			pkey_count_start=aph.fits.header.DictReader.find_max_pkey_n(hdr)
		))
		
		for i, ax in enumerate(axis_fits):
			_lgr.debug(f'{ax=}')
			aph.fits.header.set_axes_transform(hdr, 
				ax, 
				None, 
				ax_values[0,i],
				steps[i],
				result.shape[i],
				1
			)
		
		aph.fits.header.set_iwc_matrix(hdr, transform_matrix @ iwc_matrix)
		

	
	# Save the products to a FITS file
	hdu_rebinned = fits.PrimaryHDU(
		header = hdr,
		data = result
	)
	hdul_output = fits.HDUList([
		hdu_rebinned,
	])
	hdul_output.writeto(output_path, overwrite=True)
	_lgr.info(f'Written processed file to "{output_path}"')
			


def parse_args(argv):
	import os
	import aopp_deconv_tool.text
	import argparse
	
	DEFAULT_OUTPUT_TAG = '_spatialrebin'
	DESIRED_FITS_AXES = ['CELESTIAL']
	OUTPUT_COLUMNS=80
	try:
		OUTPUT_COLUMNS = os.get_terminal_size().columns - 30
	except Exception:
		pass
	
	FITS_SPECIFIER_HELP = aopp_deconv_tool.text.wrap(
		aph.fits.specifier.get_help(DESIRED_FITS_AXES).replace('\t', '    '),
		OUTPUT_COLUMNS
	)
	
	parser = argparse.ArgumentParser(
		description=__doc__, 
		formatter_class=argparse.RawTextHelpFormatter,
		epilog=FITS_SPECIFIER_HELP
	)
	
	parser.add_argument(
		'fits_spec',
		help = '\n'.join((
			f'FITS Specifier of the data to operate upon . See the end of the help message for more information',
			f'required axes: {", ".join(DESIRED_FITS_AXES)}',
		)),
		type=str,
		metavar='FITS Specifier',
	)
	parser.add_argument(
		'-o', 
		'--output_path', 
		type=FPath,
		metavar='str',
		default='{parent}/{stem}{tag}{suffix}',
		help = '\n    '.join((
			f'Output fits file path, supports keyword substitution using parts of `fits_spec` path where:',
			'{parent}: containing folder',
			'{stem}  : filename (not including final extension)',
			f'{{tag}}   : script specific tag, "{DEFAULT_OUTPUT_TAG}" in this case',
			'{suffix}: final extension (everything after the last ".")',
			'\b'
		))
	)
		
	parser.add_argument('--rebin_step', action='extend', nargs=2, type=float, metavar='float', help='step for rebinning operation in each direction, units are whatever the fits file is in, normally degrees')
	
	args = parser.parse_args(argv)
	
	args.fits_spec = aph.fits.specifier.parse(args.fits_spec, DESIRED_FITS_AXES)
		
	other_file_path = Path(args.fits_spec.path)
	args.output_path = args.output_path.with_fields(
		tag=DEFAULT_OUTPUT_TAG, 
		parent=other_file_path.parent, 
		stem=other_file_path.stem, 
		suffix=other_file_path.suffix
	)
	
	
	print('INPUT PARAMETERS')
	for k,v in vars(args).items():
		print(f'    {k} : {v}')
	print('END')
	
	return args



def go(
		fits_spec,
		output_path=None,
		rebin_step=None
	):
	"""
	Thin wrapper around `run()` to accept string inputs.
	As long as the names of the arguments to this function 
	are the same as the names expected from the command line
	we can do this programatically
	"""
	# This must be first so we only grab the arguments to the function
	fargs = dict(locals().items())
	arglist = aopp_deconv_tool.arguments.construct_arglist_from_locals(fargs, n_positional_args=1)
	
	exec_with_args(arglist)
	return

def exec_with_args(argv):
	args = parse_args(argv)
	
	
	run(
		args.fits_spec,
		args.output_path,
		args.rebin_step,	
	)

if __name__=='__main__':
	exec_with_args(sys.argv[1:])
	
	


"""
Shapes for use with geometry package
"""
from typing import Callable
import dataclasses as dc

import numpy as np

import aopp_deconv_tool.geometry as geo

@dc.dataclass(slots=True)
class GeoShape:
	"""
	Represents a shape that has an inside and outside
	"""
	metric : Callable[[np.ndarray,np.ndarray],float] = dc.field(default=geo.euclidean_metric, init=False)
	
	def __contains__(self, p : np.ndarray) -> bool:
		raise NotImplementedError


@dc.dataclass(slots=True)
class PolyShape(GeoShape):
	"""
	Represents a shape that can have a position in space
	"""
	centre : np.ndarray = dc.field(default_factory = lambda : np.zeros((2,)))


@dc.dataclass(slots=True)
class Circle(PolyShape):
	radius : float = 1
	
	@classmethod
	def of_radius(cls, radius=1):
		return cls(np.zeros((2,)), radius)
	
	def __contains__(self, p : np.ndarray) -> bool:
		return self.metric(p,self.centre) <= self.radius
	
	@property
	def diameter(self) -> float:
		return 2.0*self.radius
	
	def __str__(self):
		return f'{self.__class__.__name__.rsplit(".",1)[-1]}({self.radius})'

@dc.dataclass(slots=True)
class CompoundShape(GeoShape):
	_members : list[GeoShape] = dc.field(default_factory=list)
	_centre : np.ndarray = dc.field(default_factory = lambda : np.zeros((2,)))
	
	@property
	def centre(self) -> np.ndarray:
		return self._centre
	
	@centre.setter
	def centre(self, value:np.ndarray) -> None:
		diff = value - self._centre
		for shape in self._members:
			shape.centre += diff
	
	def __add__(self, shape : GeoShape):
		if len(self._members) == 0:
			self.centre = np.array(shape.centre)
		self._members.append(shape)
		return self
	
	def __contains__(self, p: np.ndarray) -> bool:
		for shape in self._members:
			if p in shape: return True
		return False



"""
Detects artefacts, returns a "badness map" that represents how much the algorithm thinks a particular pixel is an artefact.
"""
import sys, os
from pathlib import Path
import dataclasses as dc
from typing import Literal, Any, Type, Callable
from collections import namedtuple

import numpy as np
import scipy as sp
from astropy.io import fits

# TESTING
import skimage as ski

import matplotlib.pyplot as plt

from aopp_deconv_tool.algorithm.artefact_detection import difference_of_scale_filters, wavelet_decomposition

import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.specifier
import aopp_deconv_tool.astropy_helper.fits.header
from aopp_deconv_tool.fpath import FPath
import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.axes
import aopp_deconv_tool.numpy_helper.slice
import aopp_deconv_tool.arguments

from aopp_deconv_tool.algorithm.bad_pixels.ssa_sum_prob import ssa2d_sum_prob_map, ssa2d_deviations

from aopp_deconv_tool.py_ssa import SSA

from aopp_deconv_tool.image_processing import otsu_thresholding
from aopp_deconv_tool import algorithm
import aopp_deconv_tool.algorithm.interpolate

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')


# First one of these is the default
artefact_detection_strategies= dict(
	ssa = {
		'description' : 'Uses singular spectrum analysis (SSA) to deterimine how likely a pixel is to belong to an artefact.',
		#'target' : ssa2d_sum_prob_map,
		'target' : ssa2d_deviations,
	},
	dummy = {
		'description' : 'Is a dummy option to be used when testing',
		'target' : lambda *a, **k: None,
	}
)

artefact_detection_strategy_choices = [x for x in artefact_detection_strategies]
artefact_detection_strategy_choices_help_str = '\n\t'+'\n\t'.join(f'{k}\n\t\t{v["description"]}' for k,v in artefact_detection_strategies.items())


def generate_masks_from_thresholds(data, thresholds):
	"""
	Generator that returns multiple masks when passed an iterable of numbers
	that can be used as thresholds. The first mask will be everything <= the first
	threshold; the last mask will be everything > the last threshold; intermediate
	masks will be everything in between neighbouring threshold values.
	"""
	thresholds = np.sort([t for t in thresholds if t is not None])
	_lgr.debug(f'{thresholds=}')
	for i in range(0,len(thresholds)+1):
		if i==0:
			mask = data <= thresholds[i]
		elif i == len(thresholds):
			mask = data > thresholds[i-1]
		else:
			mask = (thresholds[i-1] < data) & (data <= thresholds[i])
		yield mask

def to_unit_range(data):
	"""
	Convert `data` to a unit range, will return the `converted_data`, `offset`, and `range` used in the conversion.
	"""
	offset = np.min(data)
	range = np.max(data) - offset
	return (data-offset)/range, offset, range

def undo_unit_range(data, offset, range):
	"""
	Given some `data` (between 0,1), an `offset` (from zero) and a `range` (max-min). Will undo a conversion to unit range.
	"""
	return (data * range) + offset

def to_dtype_range(data, dtype=np.uint16):
	"""
	Rescale data so it fits into the range of the specified `dtype`, default is `dtype=uint16`.
	"""
	final_range = np.iinfo(dtype).max
	offset = np.min(data)
	range = np.max(data) - offset
	return ((data-offset)*(final_range/range)).astype(np.uint16), offset, range, final_range

def undo_range(data, offset, range, final_range):
	"""
	Undo the coversion from one range of data to another.
	"""
	return (data.astype(float)/final_range)*range + offset

def get_ski_filter(ski_filter, undo_scaling=True, dtype=np.uint16):
	"""
	When given a scikit image filter, return a new function that will implement
	that filter on some data in a generic format, not just the one the scikit image
	filter wants.
	"""
	def new_filter(data, scale, *args, **kwargs):
		if scale == 0:
			return data
		d, o, r, rr = to_dtype_range(data, dtype)
		#plt.title('d')
		#plt.imshow(d)
		#plt.show()
		z = ski_filter(d, np.ones((scale,scale)), *args, **kwargs)
		#plt.title('z')
		#plt.imshow(z)
		#plt.show()
		if undo_scaling:
			z = undo_range(z, o, r, rr)
		return z.astype(data.dtype)
	return new_filter

def run(
		fits_spec,
		output_path,
		strategy : str,
		**kwargs : dict[str : Any]
	) -> None:
	"""
	Perform the operation associated with this module.
	"""
	
	_lgr.debug(f'{kwargs=}')
	
	strategy_callable = artefact_detection_strategies[strategy]['target']
	
	kwargs['start'] = int(kwargs['start']) if kwargs['start'] >= 0 else int(-1*kwargs['start']*kwargs['w_shape']**2)
	kwargs['stop'] = int(kwargs['stop']) if kwargs['stop'] >= 0 else int(-1*kwargs['stop']*kwargs['w_shape']**2)
	
	with fits.open(Path(fits_spec.path)) as data_hdul:
		
		
		_lgr.debug(f'{fits_spec.path=} {fits_spec.ext=} {fits_spec.slices=} {fits_spec.axes=}')
		#raise RuntimeError(f'DEBUGGING')
	
		data_hdu = data_hdul[fits_spec.ext]
		data = data_hdu.data
		
		# Allocate and set to all zeros
		badness_map = np.zeros_like(data, dtype=float)

		# Loop over the index range specified by `obs_fits_spec` and `psf_fits_spec`
		for i, idx in enumerate(nph.slice.iter_indices(data, fits_spec.slices, fits_spec.axes['CELESTIAL'])):
			_lgr.debug(f'{i=}')
			
			# Ignore data that is 99% or more NAN values
			data_nan_frac = np.count_nonzero(np.isnan(data[idx]))/data[idx].size
			_lgr.debug(f'{data_nan_frac=}')
			if data_nan_frac > 0.99:
				continue 
			
			data[idx] = algorithm.interpolate.quick_remove_nan_and_inf(data[idx])
			
			
			ssa = SSA(
				data[idx],
				w_shape = kwargs['w_shape'],
				grouping = {'mode':'elementary'}
				#grouping = {'mode':'similar_eigenvalues', 'tolerance' : 2E-1}
			)
			
			# TESTING
			#ssa.plot_ssa([0,1,2,3])
			#ssa.plot_component_slices([(0,5),(5,10),(10,20),(20,25),(25,45),(45,65),(65,75),(75,80),(80,85),(85,90),(90,95),(95,None)])
			#plt.savefig('uranus_ssa_test.png')
			#plt.show()
			#sys.exit()
			
			
			"""
			# Playing around with wavelet decomposition
			#w_data = data[idx]
			w_data = ssa2d_deviations(ssa, 25,100)
			wavelet_planes = wavelet_decomposition(w_data, 0, None, sp.ndimage.gaussian_filter, 0, False)
			#wavelet_planes = wavelet_decomposition(w_data, 1, 7, lambda x,s: sp.ndimage.rank_filter(x,s//2,s), 1, True)
			#wavelet_planes = wavelet_decomposition(w_data, 1, None, sp.ndimage.maximum_filter, 1, True)
			#wavelet_planes = wavelet_decomposition(w_data, 1, None, sp.ndimage.minimum_filter, 1, True)
			#wavelet_planes = wavelet_decomposition(w_data, 1, 4, sp.ndimage.median_filter, 1, True)
			#wavelet_planes = wavelet_decomposition(w_data, 1, None, sp.ndimage.uniform_filter, 1, True)
			
			plt.clf()
			f = plt.gcf()
			f.suptitle('wavelets components')
			n_wavelet_plots = wavelet_planes.shape[0]
			n_plots = n_wavelet_plots + 2
			nr = int(np.floor(np.sqrt(n_plots)))
			nc = int(np.ceil(n_plots / nr))
			ax = f.subplots(nr, nc).flatten()
			for i in range(n_plots):
				if i < n_wavelet_plots:
					wp = wavelet_planes[i]
					ax[i].set_title(f'wavelet component {i}\n[{np.nanmin(wp):06.2E}, {np.nanmax(wp):06.2E}]')
					ax[i].imshow(wp)
					ax[i].set_axis_off()
				elif i == n_wavelet_plots + 0:
					wp = np.sum(wavelet_planes, axis=0)
					ax[i].set_title(f'sum\n[{np.nanmin(wp):06.2E}, {np.nanmax(wp):06.2E}]')
					ax[i].imshow(wp)
					ax[i].set_axis_off()
				elif i == n_wavelet_plots + 1:
					wp = w_data - np.sum(wavelet_planes, axis=0)
					ax[i].set_title(f'data - sum\n[{np.nanmin(wp):06.2E}, {np.nanmax(wp):06.2E}]')
					ax[i].imshow(wp)
					ax[i].set_axis_off()
				# elif i == n_wavelet_plots + 2:
				# 	wp = np.std(wavelet_planes, axis=(1,2))
				# 	ax[i].set_title(f'sum over wavelet components')
				# 	ax[i].plot(wp)
				# 	ax[i].set_xlabel('wavelet component')
				# 	ax[i].set_yscale('log')
				else:
					raise RuntimeError('This should never happen')
			plt.show()
			"""
			
			
			
			
			"""
			j_count = 0
			s1,s2 = (5,75)
			plt.figure()
			plt.imshow(np.sum(ssa.X_ssa[s1:s2],axis=0))
			plt.figure()
			plt.imshow(data[idx] - np.sum(ssa.X_ssa[s1:s2],axis=0))
			plt.show()
			for j in range(s1,s2):
				_lgr.debug(f'{j=}')
				j_count += 1
				#badness_map[idx] += ssa.X_ssa[j]
				#temp = data[idx]
				temp = ssa.X_ssa[j]
				#temp, toff, tr= to_unit_range(ssa.X_ssa[j])
				
				s = 2000
				scale = 1
				ski_filter = get_ski_filter(ski.filters.rank.median, undo_scaling=True, dtype=np.uint8)
				#ski_filter = get_ski_filter(ski.filters.rank.entropy, undo_scaling=False, dtype=np.uint8)
				#r = temp
				#r = ski.filters.difference_of_gaussians(temp, 0, (scale+1)**2)
				#r = temp - sp.ndimage.uniform_filter(temp, size=scale)
				r = difference_of_scale_filters(temp, 0, scale, sp.ndimage.gaussian_filter)
				#r = difference_of_scale_filters(temp, 0, scale, ski_filter)
				#r = ski_filter(temp, scale)
				r_mean = sp.ndimage.uniform_filter(r, size=s)
				r_std = np.sqrt(sp.ndimage.uniform_filter((r - r_mean)**2, size=s))
				z = (r-r_mean)/r_std
				badness_map[idx] += z
				
				#plt.imshow(r)
				#plt.show()
				#break
			#badness_map[idx] = np.fabs(badness_map[idx]/50)
			badness_map[idx] = np.fabs(badness_map[idx]/j_count)
			continue # TESTING
			"""
			
			
			
						
			# Perform artefact detection on "background", "midground" and "foreground" separately
			thresholds = otsu_thresholding.n_exact(data[idx], 2, max_elements=10000)
			
			#for mask in generate_masks_from_thresholds(data[idx], ots):
			for mask in generate_masks_from_thresholds(data[idx], thresholds):
				
				bm = strategy_callable(
					ssa, 
					start=kwargs['start'], 
					stop=kwargs['stop'],
					mask=mask,
					size=max(data[idx].shape)//4,
				)
				
				x = np.full_like(bm, fill_value=np.nan)
				x[mask] = bm[mask]
				#plt.imshow(x)
				#plt.show()
				
				badness_map[idx] += bm
				
			
	
		hdr = data_hdu.header
		param_dict = {
			'original_file' : Path(fits_spec.path).name, # record the file we used
			'strategy' : strategy,
			**dict(('.'.join((strategy,k)),v) for k,v in kwargs.items())
		}
		#for i, x in enumerate(bad_pixel_map_binary_operations):
		#	param_dict[f'bad_pixel_map_binary_operations_{i}'] = x
		
		hdr.update(aph.fits.header.DictReader(
			param_dict,
			prefix='artefact_detection',
			pkey_count_start=aph.fits.header.DictReader.find_max_pkey_n(hdr)
		))
				

	
	# Save the products to a FITS file
	hdu_badness = fits.PrimaryHDU(
		header = hdr,
		data = badness_map
	)
	hdul_output = fits.HDUList([
		hdu_badness,
	])
	hdul_output.writeto(output_path, overwrite=True)



def parse_args(argv):
	"""
	Read command-line arguments when this module is called as a script.
	"""
	
	import aopp_deconv_tool.text
	import argparse
	
	DEFAULT_OUTPUT_TAG = '_artefactmap'
	DESIRED_FITS_AXES = ['CELESTIAL']
	OUTPUT_COLUMNS=80
	try:
		OUTPUT_COLUMNS = os.get_terminal_size().columns - 30
	except Exception:
		pass
	
	FITS_SPECIFIER_HELP = aopp_deconv_tool.text.wrap(
		aph.fits.specifier.get_help(DESIRED_FITS_AXES).replace('\t', '    '),
		OUTPUT_COLUMNS
	)
	
	parser = argparse.ArgumentParser(
		description=__doc__, 
		formatter_class=argparse.RawTextHelpFormatter,
		epilog=FITS_SPECIFIER_HELP
	)
	
	parser.add_argument(
		'fits_spec',
		help = '\n'.join((
			f'FITS Specifier of the data to operate upon . See the end of the help message for more information',
			f'required axes: {", ".join(DESIRED_FITS_AXES)}',
		)),
		type=str,
		metavar='FITS Specifier',
	)
	#parser.add_argument('-o', '--output_path', help=f'Output fits file path. By default is same as fie `fits_spec` path with "{DEFAULT_OUTPUT_TAG}" appended to the filename')
	parser.add_argument(
		'-o', 
		'--output_path', 
		type=FPath,
		metavar='str',
		default='{parent}/{stem}{tag}{suffix}',
		help = '\n    '.join((
			f'Output fits file path, supports keyword substitution using parts of `fits_spec` path where:',
			'{parent}: containing folder',
			'{stem}  : filename (not including final extension)',
			f'{{tag}}   : script specific tag, "{DEFAULT_OUTPUT_TAG}" in this case',
			'{suffix}: final extension (everything after the last ".")',
			'\b'
		))
	)
	
	
	
	parser.add_argument('--strategy', choices=artefact_detection_strategy_choices, default=artefact_detection_strategy_choices[0], help=f'Strategy to use when detecting artefacts {artefact_detection_strategy_choices_help_str}')

	ssa_group = parser.add_argument_group('ssa artefact detection', 'options for singular spectrum analysis (SSA) argument detection strategy')
	ssa_group.add_argument('--ssa.w_shape', type=int, default=10, help='Window size to calculate SSA for. Will generate `w_shape`^2 SSA components')
	ssa_group.add_argument('--ssa.start', type=float, default=-0.25, help='First SSA component to be included in artefact detection calc. Negative numbers are fractions of range')
	ssa_group.add_argument('--ssa.stop',  type=float, default=-0.75, help='Last SSA component to be included in artefact detection calc. Negative numbers are fractions of range')

	args = parser.parse_args(argv)
	
	args.fits_spec = aph.fits.specifier.parse(args.fits_spec, DESIRED_FITS_AXES)
	
	#if args.output_path is None:
	#	args.output_path =  (Path(args.fits_spec.path).parent / (str(Path(args.fits_spec.path).stem)+DEFAULT_OUTPUT_TAG+str(Path(args.fits_spec.path).suffix)))
	other_file_path = Path(args.fits_spec.path)
	args.output_path = args.output_path.with_fields(
		tag=DEFAULT_OUTPUT_TAG, 
		parent=other_file_path.parent, 
		stem=other_file_path.stem, 
		suffix=other_file_path.suffix
	)
	
	args.strategy_args = dict((k[len(args.strategy)+1:],v) for k,v in vars(args).items() if k.startswith(args.strategy))	
	
	
	return args


def go(
		fits_spec,
		output_path=None,
		strategy=None,
		**kwargs
	):
	"""
	Thin wrapper around `run()` to accept string inputs.
	As long as the names of the arguments to this function 
	are the same as the names expected from the command line
	we can do this programatically
	"""
	# Add stuff to kwargs here if needed
	kwargs['ssa.w_shape'] = None
	kwargs['ssa.start'] = None
	kwargs['ssa.stop'] = None
	
	# This must be first so we only grab the arguments to the function
	fargs = dict(locals().items())
	arglist = aopp_deconv_tool.arguments.construct_arglist_from_locals(fargs, n_positional_args=1)

	exec_with_args(arglist)

def exec_with_args(argv):
	"""
	Read arguments and run the module
	"""
	args = parse_args(argv)
	
	run(
		args.fits_spec, 
		output_path=args.output_path, 
		strategy=args.strategy, 
		**args.strategy_args
	)

if __name__ == '__main__':
	exec_with_args(sys.argv[1:])


"""
Contains classes that assist with plotting histograms.
"""

import sys
from time import sleep
from typing import Any
import dataclasses as dc


import numpy as np
import matplotlib as mpl
# NOTE: If failing, try turing this back on
#mpl.use('TKAgg')
import matplotlib.pyplot as plt

import aopp_deconv_tool.plot_helper as plot_helper
from aopp_deconv_tool.plot_helper.base import Base, AxisDataMapping


@dc.dataclass(repr=False, eq=False, slots=True)
class PlotSet:
	"""
	Defines a set of plots that can be animated
	"""
	fig : mpl.figure.Figure
	title : str = 'Plot Set'
	plots : list[Base] = dc.field(default_factory=list)
	blit : bool = True
	cadence : int = 1
	
	# Internal attributes
	n_frames : int = dc.field(default=0, init=False)
	blit_regions : Any = dc.field(default=None, init=False)
	blit_regions_data : Any = dc.field(default=None, init=False)
	title_hdl : Any = dc.field(default=None, init=False)
	title_blit_region : Any = dc.field(default=None, init=False)
	title_blit_region_data : Any = dc.field(default=None, init=False)
		
	def attach_datasource(self, datasource : Any):
		"""
		loop through plots, any with no attached datasource will use this one
		"""
		for p in self.plots:
			if p.is_datasource_attached():
				continue
			p.attach_datasource(datasource, p.datasource_data_getter)
		return self
		
	def show(self):
		self.title_hdl = self.fig.suptitle(self.title.format(self=self))
		self.n_frames = 0
		self.fig.canvas.draw()
		plt.show(block=False)
		
		#if self.blit:
			#plt.pause(0.4)
			#self.blit_regions = [self.fig.canvas.copy_from_bbox(ax.bbox) for ax in self.fig.axes]
			#self.blit_regions = [self.fig.canvas.copy_from_bbox(ax.get_tightbbox()) for ax in self.fig.axes]
	
	
	def get_ax_bb(self, ax):
		return ax.get_tightbbox().expanded(1.2,1.2)
	
	def update(self):
		
		if self.blit:
			# do one full draw then animate with blit
			if self.n_frames == 0:
				
				for ax in self.fig.axes:
					ax.drawn = False
				
				for plot in self.plots:
					plot.update()
				plt.pause(0.1)
					
				# set axes invisible that will change
				self.title_hdl.set_visible(False)
				for p in self.plots:
					p.ax.set_visible(True if p.static_frame else False)
				
					
				self.fig.canvas.draw()
				self.fig.canvas.flush_events()
				
				# set axes visible again
				self.title_hdl.set_visible(True)
				for p in self.plots:
					p.ax.set_visible(True)
					
				# need axes visible to get their extent
				self.title_blit_region = self.title_hdl.get_window_extent().expanded(1.2,1.2)
				self.blit_regions = tuple(self.get_ax_bb(ax) for ax in self.fig.axes)
				
				# set axes invisible that will change
				self.title_hdl.set_visible(False)
				for p in self.plots:
					p.ax.set_visible(True if p.static_frame else False)
				
				self.fig.canvas.draw()
				self.fig.canvas.flush_events()
						# need axes invisible to get the blank background
				self.title_blit_region_data = self.fig.canvas.copy_from_bbox(self.title_blit_region)
				self.blit_regions_data = tuple(self.fig.canvas.copy_from_bbox(r) for r in self.blit_regions)
						
				# set axes visible again
				self.title_hdl.set_visible(True)
				for p in self.plots:
					p.ax.set_visible(True)
				
				self.fig.canvas.draw()
				self.fig.canvas.flush_events()
				
				
				self.n_frames += 1
				return
				
			if self.n_frames % self.cadence == 0 :	
				# remove last frame plots
				self.fig.canvas.restore_region(self.title_blit_region_data)
				for ax, blit_region_data in zip(self.fig.axes, self.blit_regions_data):
					self.fig.canvas.restore_region(blit_region_data)
					ax.drawn=False
			
				# blit new frame plots
			self.title_hdl.set_text(self.title.format(self=self))
			
			if self.n_frames % self.cadence == 0 :
				self.fig.draw_artist(self.title_hdl)
				self.fig.canvas.blit(self.title_blit_region)
			
			for plot in self.plots:
				plot.update()
				if self.n_frames % self.cadence == 0 :
					self.fig.canvas.blit(self.get_ax_bb(plot.ax))
		else:
			self.title_hdl = self.fig.suptitle(self.title.format(self))
			for plot in self.plots:
				plot.update()
			
			if self.n_frames % self.cadence == 0 :
				self.fig.canvas.draw()
		
		if self.n_frames % self.cadence == 0 :	
			self.fig.canvas.flush_events()
		
	
		self.n_frames += 1


@dc.dataclass(repr=False, eq=False, slots=True)
class Histogram(Base):
	title : str = 'Histogram'
	datasource_name : str = 'histogram datasource'
	
	# Order corresponds to which axes they are associated with (0,1,2,..) -> (x,y,z,...)
	axis_data_mappings : tuple[AxisDataMapping] = dc.field(default_factory= lambda :(
		AxisDataMapping('value','bins',limit_getter=plot_helper.LimRememberExtremes()),
		AxisDataMapping('count','_hist',limit_getter=plot_helper.LimRememberExtremes()))
	)
	nbins : int = 100
	
	# Internal
	_hist : np.ndarray = None
	_bins : np.ndarray = None
	
	
	@property
	def bins(self):
		return self._bins
	
	@bins.setter
	def bins(self, value):
		self._bins[:-1] = value
		self._bins[-1] = self._bins[-2] + (self._bins[-2] - self.bins[-3])
	
	@property
	def hist(self):
		return self._hist
	
	@hist.setter
	def hist(self, value):
		self._hist[1:-1] = value
	
	def update_plot_data(self, data):
		self.hist[1:-1], self.bins = np.histogram(data, bins=self.nbins)
		self.hdl.set_data(self.bins, self.hist)
	
	def on_attach_datasource(self):
		super(Histogram, self).on_attach_datasource()
		# initialise data holders, need to have zeros at either end due to how `self.ax.step` works
		self._hist = np.zeros((self.nbins+2,))
		self._bins = np.linspace(0,1,self.nbins+2,endpoint=True)

	def on_attach_ax(self):
		super(Histogram, self).on_attach_ax()
		self.hdl = self.ax.step([],[], label=self.datasource_name, **self.plt_kwargs)[0]
	

@dc.dataclass(repr=False, eq=False, slots=True)
class Image(Base):
	title : str = 'Image'
	datasource_name : str = 'image datasource'
	
	# Order corresponds to which axes they are associated with (0,1,2,..) -> (x,y,z,...)
	axis_data_mappings : tuple[AxisDataMapping] = dc.field(default_factory= lambda :(
		AxisDataMapping('x',None), 
		AxisDataMapping('y',None), 
		AxisDataMapping('brightness', '_z_data'))
	)
	
	_z_data : np.ndarray = None
	
	def update_plot_data(self, data):
		
		self._z_data = data
		if self.hdl is None:
			#self.x_mesh, self.y_mesh = np.indices(self._z_data.shape) #np.arange(0,data.shape[0],1)
			#self.y_mesh = np.arange(0,data.shape[1],1)
			#self.hdl = self.ax.pcolormesh(self.x_mesh, self.y_mesh, self._z_data, label=self.datasource_name, shading='nearest', **self.plt_kwargs)
			kwargs = dict(origin='lower', interpolation='none')
			kwargs.update(self.plt_kwargs)
			self.hdl = self.ax.imshow(self._z_data, label=self.datasource_name, **kwargs)
			#self.hdl = self.ax.imshow(np.full_like(self._z_data, fill_value=0), label=self.datasource_name, **kwargs)
			#self.hdl = self.ax.imshow(np.zeros_like(self._z_data), label=self.datasource_name, **kwargs)
			#self.ax.legend()
		else:
			self.hdl.set_data(self._z_data)
	
	def on_attach_datasource(self):
		super(Image, self).on_attach_datasource()

	def on_attach_ax(self):
		super(Image, self).on_attach_ax()


@dc.dataclass(repr=False, eq=False, slots=True)
class VerticalLine(Base):
	title : str = 'Vertical Line'
	datasource_name : str = 'vertical line datasource'
	
	# Order corresponds to which axes they are associated with (0,1,2,..) -> (x,y,z,...)
	axis_data_mappings : tuple[AxisDataMapping] = dc.field(default_factory= lambda :(AxisDataMapping(None,None),))
	
	
	# Internal
	_x_pos : float = None
	
	def update_plot_data(self, data):
		self._x_pos = data
		if self.hdl is not None:
			self.hdl.remove()
		self.hdl = self.ax.axvline(self._x_pos, label=self.datasource_name, **self.plt_kwargs)
	
	def on_attach_datasource(self):
		super(VerticalLine, self).on_attach_datasource()

	def on_attach_ax(self):
		super(VerticalLine, self).on_attach_ax()


@dc.dataclass(repr=False, eq=False, slots=True)
class HorizontalLine(Base):
	title : str = 'Horizontal Line'
	datasource_name : str = 'horizontal line datasource'
	
	# Order corresponds to which axes they are associated with (0,1,2,..) -> (x,y,z,...)
	axis_data_mappings : tuple[AxisDataMapping] = dc.field(default_factory= lambda :(AxisDataMapping(None,None),))
	
	
	# Internal
	_y_pos : float = None
	
	def update_plot_data(self, data):
		
		self._y_pos = data
		#print(f'HorizontalLine {data=}')
		if self.hdl is not None:
			self.hdl.remove()
		self.hdl = self.ax.axhline(self._y_pos, label=self.datasource_name, **self.plt_kwargs)
	
	def on_attach_datasource(self):
		super(HorizontalLine, self).on_attach_datasource()

	def on_attach_ax(self):
		super(HorizontalLine, self).on_attach_ax()


@dc.dataclass(repr=False, eq=False, slots=True)
class IterativeLineGraph(Base):
	title : str = 'Iterative Line Graph'
	datasource_name : str = 'iterative line datasource'
	
	axis_data_mappings : tuple[AxisDataMapping] = dc.field(default_factory= lambda : (
			AxisDataMapping(
				'iteration',
				'_x',
				limit_getter=plot_helper.LimRememberExtremes(plot_helper.LimFixed(0,None))
			),
			AxisDataMapping(
				'value',
				'_y',
				limit_getter=plot_helper.LimRememberExtremes(plot_helper.LimFixed(0,None))
			)
		)
	)
	
	_x : list = dc.field(default_factory=list)
	_y : list = dc.field(default_factory=list)
	
	def update_plot_data(self, data):
		#print(f'{self.datasource_name} {data=}')
		self._x.append(len(self._x))
		self._y.append(data)
		self.hdl.set_data(self._x,self._y)
	
	def on_attach_datasource(self):
		super(IterativeLineGraph, self).on_attach_datasource()
		self._x = []
		self._y = []

	def on_attach_ax(self):
		super(IterativeLineGraph, self).on_attach_ax()
		self.hdl = self.ax.plot(self._x, self._y, label=self.datasource_name, **self.plt_kwargs)[0]



@dc.dataclass(repr=False, eq=False, slots=True)
class IterativeLogLineGraph(IterativeLineGraph):
	title : str = 'Iterative Log Line Graph'
	datasource_name : str = 'iterative log line datasource'
	
	ax_funcs=[lambda ax: ax.set_yscale('log')]



"""
Holds a representation of optical components that can be combined together
into a set of optical components that describe an optical system.
"""

import dataclasses as dc
from typing import Literal, Any

import numpy as np

import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.array
from aopp_deconv_tool.numpy_helper.array import S,N
from aopp_deconv_tool.geometry.shape import GeoShape, Circle
from aopp_deconv_tool.optics.geometric.light_beam import LightBeam, LightBeamSet

def sign(a):
	return -1 if a < 0 else 1



@dc.dataclass(slots=True)
class OpticalComponent:
	"""
	Represents a component of an optical system. As I am working in "optical
	distance of the undeviated ray", there are only 3 basic things to model. 
	Apertures, obstructions, and reflector/refractors (which have the same 
	behaviour on optical distance).
	"""
	position : float = 0 # in units of "optical distance of undeviated ray"
	name : str = 'optical_component' # Name of this component
	_is_aperture_stop : bool = dc.field(default=False, init=False, repr=False, compare=False)
	
	
	def out_light_beam(self, in_light_beam : LightBeam) -> LightBeam:
		"""
		For an incident light beam `in_light_beam`, calculate the parameters of 
		a beam after passing through this component.
		"""
		raise NotImplementedError
	
	def __str__(self) -> str:
		"""
		A textual representation of this component
		"""
		raise NotImplementedError

@dc.dataclass(slots=True)
class Refractor(OpticalComponent):
	"""
	A refractor or reflector (as position is in "optical distance of undeviated
	ray")
	"""
	shape : GeoShape = dc.field(default_factory=lambda: Circle.of_radius(8)) # meters. Cross-sectional shape of the light in the path
	focal_length : float = 120 # meters. Distance in the direction of the optical path that this refractor will focus light to.
	
	def out_light_beam(self, in_light_beam : LightBeam) -> LightBeam:
		# Gradient of the high and low edges of the beam are adjusted
		# as they now focus to a new position.
		w_a_pos = in_light_beam.w_a(self.position)
		w_b_pos = in_light_beam.w_b(self.position)
		
		# Need to know if this aperture is the aperture stop or not
		print(f'{self.shape=}')
		if (sign(w_a_pos)*(w_b_pos)) <= (self.shape.radius):
			self._is_aperture_stop=False
			c_b = w_b_pos
		else:
			self._is_aperture_stop=True
			c_b = sign(w_a_pos)*self.shape.radius
			
		c_a = min(sign(w_a_pos)*(w_a_pos), self.shape.radius)*sign(w_a_pos)
		return LightBeam(
			c_a, 
			in_light_beam.m_a-c_a/self.focal_length, 
			c_b, 
			in_light_beam.m_b-c_b/self.focal_length, 
			self.position
		)
	
	def __str__(self):
		return f'{self.__class__.__name__.rsplit(".",1)[-1]}({self.name},{self.position},{str(self.shape)},focal_length={self.focal_length})'

@dc.dataclass(slots=True)
class Aperture(OpticalComponent):
	"""
	A transparent opening surrounded by opaque material. Can also model flat mirrors
	"""
	shape : GeoShape = dc.field(default_factory=lambda: Circle.of_radius(2.52/18)) # NOTE: This number is in fraction of pupil exit diameter.
	
	def out_light_beam(self, in_light_beam : LightBeam) -> LightBeam:
		# high edge of the light beam should be adjusted to exclude the occluded region
		# gradient of the high edge of the beam adjusted as it should focus in the
		# same place as before.
		w_a_pos = in_light_beam.w_a(self.position)
		w_b_pos = in_light_beam.w_b(self.position)
		if (sign(w_a_pos)*w_a_pos) >= self.shape.radius and (sign(w_b_pos)*w_b_pos) >= self.shape.radius:
			# We have no light that can get through the aperture, return null beam
			return LightBeam(np.nan,np.nan,np.nan,np.nan,self.position)
			
		w_b_pos = in_light_beam.w_b(self.position)
		w_pos = w_b_pos - w_a_pos
		
		# Need to know if this aperture is the aperture stop or not
		if sign(w_a_pos)*(w_b_pos) <= self.shape.radius:
			self._is_aperture_stop=False
			c_b = w_b_pos
		else:
			self._is_aperture_stop=True
			c_b = sign(w_a_pos)*self.shape.radius
		
		c_a = min(sign(w_a_pos)*(w_a_pos), self.shape.radius)*sign(w_a_pos)
		w_b_pos_frac = (c_b-w_a_pos) / w_pos
		w_a_pos_frac = (c_a-w_a_pos) / w_pos
		
		m_b_pos = (in_light_beam.m_b - in_light_beam.m_a)*w_b_pos_frac + in_light_beam.m_a
		m_a_pos = (in_light_beam.m_b - in_light_beam.m_a)*w_a_pos_frac + in_light_beam.m_a
		
		lb= LightBeam(
			c_a,
			m_a_pos,
			c_b, 
			m_b_pos,
			self.position
		)
		print(f'{lb=}')
		return lb
	
	def __str__(self):
		return f'{self.__class__.__name__.rsplit(".",1)[-1]}({self.name},{self.position},{str(self.shape)})'


@dc.dataclass(slots=True)
class Obstruction(OpticalComponent):
	"""
	Opaque material surrounded by transparent material. Used to model e.g., the back of secondary mirrors.
	"""
	shape : GeoShape = dc.field(default_factory=lambda: Circle.of_radius(2.52/18)) # NOTE: This number is in fraction of pupil exit diameter.
	
	def out_light_beam(self, in_light_beam : LightBeam) -> LightBeam:
		# low edge of beam must be adjusted to exclude the occluded part,
		# the gradient of the low edge of the beam should be adjusted to
		# ensure it still focuses at the same point as before
		w_a_pos = in_light_beam.w_a(self.position)
		w_b_pos = in_light_beam.w_b(self.position)
		if (sign(w_b_pos)*w_a_pos) <= self.shape.radius and (sign(w_b_pos)*w_b_pos) <= self.shape.radius:
			# We have no light that can get through the aperture, return null beam
			return LightBeam(np.nan,np.nan,np.nan,np.nan,self.position)
			
		w_b_pos = in_light_beam.w_b(self.position)
		w_pos = w_b_pos - w_a_pos
		
		c_b = max(sign(w_b_pos)*(w_b_pos), self.shape.radius)*sign(w_b_pos)
		c_a = max(sign(w_b_pos)*(w_a_pos), self.shape.radius)*sign(w_b_pos)
		w_b_pos_frac = (c_b-w_a_pos) / w_pos
		w_a_pos_frac = (c_a-w_a_pos) / w_pos
		
		m_b_pos = (in_light_beam.m_b - in_light_beam.m_a)*w_b_pos_frac + in_light_beam.m_a
		m_a_pos = (in_light_beam.m_b - in_light_beam.m_a)*w_a_pos_frac + in_light_beam.m_a
		
		return LightBeam(
			c_a, 
			m_a_pos,
			c_b,
			m_b_pos,
			self.position
		)
	
	def __str__(self):
		return f'{self.__class__.__name__.rsplit(".",1)[-1]}({self.name},{self.position},{str(self.shape)})'



@dc.dataclass(slots=True)
class OpticalComponentSet:
	"""
	A set of optical components that make up an optical system. The components 
	are placed in order of the "optical distance of the undeviated ray". I.e.
	an [aperture]->[mirror] system, will be modelled as 
	[aperture]->[refractor]->[aperture] as the undeviated ray encounters the 
	[aperture] twice. This also enables refractors and reflectors to be modelled
	as the same thing.
	"""
	_optical_path : list[OpticalComponent] = dc.field(default_factory=list, init=False, repr=False, compare=False) # the ordered list of components in the optical path, position is in terms of "optical distance of undeviated ray"
	_aperture_stop_idx : int | None = None # index of the component that is the aperture stop of the system. If None, then aperture stop has not been determined
	
	_lbs : Any = dc.field(default=None, init=False, repr=False, compare=False)

	@classmethod
	def from_components(cls, optical_components : list[OpticalComponent]):
		"""
		Example:
			OpticalComponentSet.from_components([
				Aperture(0, 'objective aperture', Circle.of_radius(7)), 
				Obstruction(3, 'secondary mirror back', Circle.of_radius(2)), 
				Refractor(100, 'primary mirror', Circle.of_radius(7), 120),
				Aperture(197, 'secondary mirror front', circle.of_radius(2))
			])
		"""
		ocs = cls()
		for oc in optical_components:
			print(oc)
			ocs.insert(oc)
		return ocs
	
	def get_evaluation_positions(self, delta=1E-5):
		"""
		When plotting all light beams in a system, want to have positions of 
		beam envelope before and after each optical component.
		"""
		x = np.ndarray((2*len(self)+1,))
		xmax = -np.inf
		for i, oc in enumerate(self._optical_path):
			_xmax = oc.position + oc.focal_length if hasattr(oc,'focal_length') else 0
			if xmax < _xmax : xmax = _xmax
			x[2*i] = oc.position-delta
			x[2*i+1] = oc.position+delta
		x[-1] = xmax
		return x
			
	
	def calc_full_light_beam(self):
		if self._lbs is None:
			self._lbs = self.get_light_beam(LightBeam(0,0,np.inf,0,-1))
		
	
	
	def get_pupil_function_scale(self, expansion_factor):
		self.calc_full_light_beam()
		lb = self._lbs[-1]
		w_max = max(abs(lb.c_a),abs(lb.c_b)) # this is a radius
		scale = ((2*w_max)*expansion_factor, (2*w_max)*expansion_factor) # need a diameter
		return scale
	
	def pupil_function(self, 
			shape=(101,101), 
			scale=None,
			expansion_factor = 1.1,
			supersample_factor = 1
		):
		"""
		Calculates the pupil function at the last optical component.
		
		Returned scale is in units of length (normally meters)
		"""
		
		
		self.calc_full_light_beam()
		# get the final light beam
		lb = self._lbs[-1]
		print(f'{lb=}')
		pf_pos = lb.o
		if scale is None:
			scale = self.get_pupil_function_scale(expansion_factor)
		else:
			scale = tuple(s*expansion_factor for s in scale)
		print(f'{scale=}')
		
		# assume cylindrically symmetric
		centre_offsets = nph.array.offsets_from_point(
			tuple(int(s*expansion_factor*supersample_factor) for s in shape),
			None, 
			np.array(scale,dtype=float)
		)
		print(f'{centre_offsets}')
		pf_rad = np.sqrt(np.sum(centre_offsets**2, axis=0))
		
		
		lb = self._lbs.get_light_beam_at_position(pf_pos)	
		w_a, w_b = lb(pf_pos) if lb is not None else (np.nan, np.nan)
		
		print(f'{w_a=} {w_b=}')
		pf_val = np.zeros_like(pf_rad, dtype=float)
		pf_val[pf_rad < (w_b)] = 1
		pf_val[pf_rad < (w_a)] = 0
		print(pf_rad)
		print(pf_val)
		return pf_val
	
	def psf(self, 
			shape=(101,101), 
			scale = None,
			expansion_factor = 1.1,
			supersample_factor = 1
		):
		"""
		Calculates the PSF.
		
		Returned scale is in radians/wavelength, multiply by light wavelength
		to get the PSF in radians.
		"""
		# NOTE: Not accounting for wavelength of light in working out the PSF
		scale, pf = self.pupil_function(self, shape, scale, expansion_factor, supersample_factor)
		pf_fft = np.fft.fftshift(np.fft.fftn(pf))
		psf = (np.conj(pf_fft)*pf_fft)
		return scale, psf
	
	def otf(self, 
			shape=(101,101), 
			scale=None,
			expansion_factor = 1.1,
			supersample_factor = 1
		):
		"""
		Gets the optical transfer function.
		
		Returned scale is in UNKNOWN AT THIS TIME
		"""
		scale, psf = self.psf(shape, scale, expansion_factor, supersample_factor)
		otf = np.fft.fftshift(np.fft.fftn(np.fft.ifftshift(psf)))
		return scale, otf
	
	def _component_name_present(self, name):
		"""
		returns True if a component of `name` is present in the set
		"""
		for oc in self._optical_path:
			if oc.name == name:
				return True
		return False
	
	
	def _do_insert(self, idx : int, oc : OpticalComponent):
		"""
		Perform insertion of optical component `oc` at positoin `idx`
		"""
		self._optical_path.insert(idx, oc)
	
	def insert(self, oc : OpticalComponent):
		"""
		Insert an optical component `oc` based upon the value of `oc.position`.
		"""
		inserted = False
		if self._component_name_present(oc.name):
			raise RuntimeError(f'A component with the name "{oc.name}" already exists in the component set.')
		
		print(f'{[x.position for x in self._optical_path]}')
		for i in range(len(self._optical_path)):
			print(i, self._optical_path[i].position, oc.position)
			if self._optical_path[i].position > oc.position:
				self._do_insert(i, oc)
				inserted=True
		if not inserted:
			self._do_insert(len(self._optical_path), oc)
	
	def get_component_before_pos(self, pos) -> OpticalComponent:
		"""
		For some `pos` (units are "optical distance of undeviated ray"), find the
		component that is before this position.
		"""
		prev = None
		for oc in self:
			if oc.position < pos:
				prev = oc
			else:
				break
		if prev is not None:
			return prev
		raise RuntimeError(f'No optical components before position {pos}, first component is at position {self._optical_path[0]}')
	
	def get_component_index_before_pos(self, pos) -> int:
		"""
		For some `pos` (units are "optical distance of undeviated ray"), find the
		index of the component that is before this position.
		"""
		prev = None
		for i, oc in enumerate(self):
			if oc.position < pos:
				prev = i
			else:
				break
		if prev is not None:
			return prev
		raise RuntimeError(f'No optical components before position {pos}, first component is at position {self._optical_path[0]}')
	
	def get_components_by_class(self, aclass):
		"""
		Return all components in this set that are of `aclass`
		"""
		for oc in enumerate(self._optical_path):
			if isinstance(oc, aclass):
				yield oc
	
	def __len__(self):
		"""
		Return the number of components in this set
		"""
		return len(self._optical_path)
	
	def __iter__(self):
		"""
		Get an iterator over the components in this set
		"""
		return iter(self._optical_path)
	
	def __getitem__(self, idx : str | int):
		"""
		Get a component in this set by either index, or name.
		"""
		t_idx = type(idx)
		if t_idx == str:
			for oc in self._optical_path:
				if oc.name == idx:
					return oc
			return None
		if t_idx == int:
			return self._optical_path[idx]
		
		raise IndexError(f'Cannot index an OpticalComponentSet with an index of type {type(idx)}')
	
	def get_light_beam(self, in_light_beam : LightBeam) -> LightBeamSet:
		"""
		Get the set of light beams that result from an incident light beam 
		`in_light_beam`. This set of light beams respresents the geometric 
		optics approximation of light's behaviour in the system.
		"""
		lb_list = [in_light_beam]
		for i, oc in enumerate(self._optical_path):
			lb_list.append(oc.out_light_beam(lb_list[i]))
			
			# Later aperture stops in the system replace earlier ones.
			if oc._is_aperture_stop:
				self._aperture_stop_idx = i
		return LightBeamSet(lb_list)

	def get_aperture_stop(self):
		if self._aperture_stop_idx is None:
			raise RuntimeError("Aperture stop has not been determined yet, run method 'self.get_light_beam(...)' first")
		return self._optical_path[self._aperture_stop_idx]

 
